# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for DebusineTaskConfiguration collections."""

import unittest
from typing import Any, ClassVar, override
from unittest import mock
from unittest.mock import sentinel

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts.models import (
    BareDataCategory,
    CollectionCategory,
    DebusineTaskConfiguration,
    TaskTypes,
)
from debusine.db.models import Collection, CollectionItem
from debusine.db.playground import scenarios
from debusine.db.tasks import DBTask
from debusine.server.collections import DebusineTaskConfigurationManager
from debusine.server.collections.base import ItemAdditionError
from debusine.server.collections.debusine_task_configuration import (
    Changes,
    ConfigurationLookup,
    apply_configuration,
    build_configuration,
)
from debusine.tasks import TaskConfigError
from debusine.tasks.models import SbuildData, SbuildInput
from debusine.test.django import TestCase


class TaskConfigTestCase(TestCase):
    """Base class for tests that apply task configuration."""

    scenario = scenarios.DefaultContext()
    collection: ClassVar[Collection]
    manager: ClassVar[DebusineTaskConfigurationManager]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.collection = cls.scenario.default_task_configuration_collection
        cls.manager = DebusineTaskConfigurationManager(
            collection=cls.collection
        )

    @classmethod
    def _add_config(cls, entry: DebusineTaskConfiguration) -> CollectionItem:
        """Add a config entry to config_collection."""
        return cls.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=cls.scenario.user,
            data=entry,
        )


class DebusineTaskConfigurationManagerTests(TaskConfigTestCase):
    """Tests for DebusineTaskConfigurationManager."""

    def test_add_bare_data_no_data(self) -> None:
        """`add_bare_data` requires item data."""
        with self.assertRaisesRegex(
            ItemAdditionError,
            "Adding to debusine:task-configuration requires data",
        ):
            self.manager.add_bare_data(
                BareDataCategory.TASK_CONFIGURATION,
                user=self.scenario.user,
            )

    def test_add_bare_data_refuses_name(self) -> None:
        """`add_bare_data` refuses an explicitly-specified name."""
        with self.assertRaisesRegex(
            ItemAdditionError,
            "Cannot use an explicit item name when adding to "
            "debusine:task-configuration",
        ):
            self.manager.add_bare_data(
                BareDataCategory.TASK_CONFIGURATION,
                user=self.scenario.user,
                data=DebusineTaskConfiguration(
                    task_type=TaskTypes.WORKER,
                    task_name="sbuild",
                    subject="base-files",
                    context="sid",
                ),
                name="override",
            )

    def test_add_bare_data_raise_item_addition_error(self) -> None:
        """`add_bare_data` raises an error for duplicate names."""
        data = DebusineTaskConfiguration(template="test")
        self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
        )

        with self.assertRaisesRegex(
            ItemAdditionError, "db_collectionitem_unique_active_name"
        ):
            self.manager.add_bare_data(
                BareDataCategory.TASK_CONFIGURATION,
                user=self.scenario.user,
                data=data,
            )

    def test_add_bare_data_replace(self) -> None:
        """`add_bare_data` can replace an existing bare data item."""
        data = DebusineTaskConfiguration(template="test")

        item_old = self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
        )

        item_new = self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
            replace=True,
        )

        serialized_data = data.model_dump(mode="json", exclude_unset=True)

        item_old.refresh_from_db()
        self.assertEqual(item_old.name, "template:test")
        self.assertEqual(item_old.child_type, CollectionItem.Types.BARE)
        self.assertEqual(item_old.data, serialized_data)
        self.assertEqual(item_old.removed_by_user, self.scenario.user)
        self.assertIsNotNone(item_old.removed_at)
        self.assertEqual(item_new.name, "template:test")
        self.assertEqual(item_new.child_type, CollectionItem.Types.BARE)
        self.assertEqual(item_new.data, serialized_data)
        self.assertIsNone(item_new.removed_at)

    def test_add_bare_data_replace_nonexistent(self) -> None:
        """Replacing a nonexistent bare data item is allowed."""
        data = DebusineTaskConfiguration(template="test")

        item = self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
            replace=True,
        )

        self.assertEqual(item.name, "template:test")
        self.assertEqual(item.child_type, CollectionItem.Types.BARE)
        self.assertEqual(
            item.data, data.model_dump(mode="json", exclude_unset=True)
        )

    def test_remove_item_bare_data(self) -> None:
        """`remove_item` removes a bare data item."""
        data = DebusineTaskConfiguration(template="test")
        item = self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
        )

        self.manager.remove_item(item, user=self.scenario.user)

        self.assertEqual(item.removed_by_user, self.scenario.user)
        self.assertIsNotNone(item.removed_at)

    def test_lookup_unexpected_format(self) -> None:
        """`lookup` raises `LookupError` for an unexpected format."""
        with self.assertRaisesRegex(
            LookupError, r'^Unexpected lookup format: "foo:bar"'
        ):
            self.manager.lookup("foo:bar")

    def test_lookup_return_none(self) -> None:
        """`lookup` returns None if there are no matches."""
        self.assertIsNone(self.manager.lookup("name:does-not-exist"))

    def test_lookup_by_name(self) -> None:
        """Lookup by name works."""
        data = DebusineTaskConfiguration(template="test")
        item = self.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=data,
        )
        self.assertEqual(self.manager.lookup("name:template:test"), item)


class ConfigurationLookupTests(TaskConfigTestCase):
    """Tests for :py:class:`ConfigurationLookup`."""

    def lookup(
        self,
        task_type: TaskTypes = TaskTypes.WORKER,
        task_name: str = "noop",
        subject: str | None = None,
        context: str | None = None,
        tags_provided: set[str] | None = None,
        tags_required: set[str] | None = None,
    ) -> ConfigurationLookup:
        """Return the ConfigurationLookup to test."""
        lookup = ConfigurationLookup(self.collection)
        lookup.set_task_type_name_constraints(task_type, task_name)
        lookup.set_subject_context_constraints(subject, context)
        lookup.set_tags_provided(tags_provided or set())
        lookup.set_tags_required(tags_required or set())
        return lookup

    def test_set_task(self) -> None:
        # Build a mock task
        task = mock.MagicMock(spec=DBTask)
        task.TASK_TYPE = sentinel.type
        task.name = sentinel.name
        task.get_task_configuration_subject_context = mock.MagicMock(
            return_value=(sentinel.subject, sentinel.context)
        )
        provided = ttags.TaskProvidedTags()
        provided.add(ttags.ProvenanceProvided.SYSTEM, {"test", "provided"})
        required = wtags.TaskRequiredTags()
        required.add(wtags.ProvenanceRequired.SYSTEM, {"test", "required"})
        task.compute_scheduler_tags = mock.MagicMock(
            return_value=(provided, required)
        )

        lookup = ConfigurationLookup(self.collection)
        with (
            self.spy_object(lookup, "set_task_type_name_constraints") as set_tn,
            self.spy_object(
                lookup, "set_subject_context_constraints"
            ) as set_sc,
            self.spy_object(lookup, "set_tags_provided") as set_tp,
            self.spy_object(lookup, "set_tags_required") as set_tr,
        ):
            lookup.set_task(task)

        set_tn.assert_called_once_with(sentinel.type, sentinel.name)
        set_sc.assert_called_once_with(sentinel.subject, sentinel.context)
        set_tp.assert_called_once_with({"test", "provided"})
        set_tr.assert_called_once_with({"test", "required"})

    def test_candidates_annotations(self) -> None:
        self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop"
            )
        )
        self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop", subject="subject"
            )
        )
        self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop", context="context"
            )
        )
        self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="provided",
                tags_provided=["a", "b"],
            )
        )
        self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="required",
                tags_required=["c", "d"],
            )
        )
        self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="exprs",
                tags_provided="a or b",
                tags_required="c or d",
            )
        )
        with (
            mock.patch(
                "debusine.server.collections.debusine_task_configuration.ConfigurationLookup.set_subject_context_constraints"
            ),
            mock.patch(
                "debusine.server.collections.debusine_task_configuration.ConfigurationLookup.set_tags_provided"
            ),
            mock.patch(
                "debusine.server.collections.debusine_task_configuration.ConfigurationLookup.set_tags_required"
            ),
        ):
            items = list(self.lookup().candidates.order_by("id"))

        self.assertEqual(items[0].tags_provided, [])  # type: ignore[attr-defined]
        self.assertEqual(items[0].tags_required, [])  # type: ignore[attr-defined]
        self.assertIsNone(items[0].subject)  # type: ignore[attr-defined]
        self.assertIsNone(items[0].context)  # type: ignore[attr-defined]

        self.assertEqual(items[1].tags_provided, [])  # type: ignore[attr-defined]
        self.assertEqual(items[1].tags_required, [])  # type: ignore[attr-defined]
        self.assertEqual(items[1].subject, "subject")  # type: ignore[attr-defined]
        self.assertIsNone(items[1].context)  # type: ignore[attr-defined]

        self.assertEqual(items[2].tags_provided, [])  # type: ignore[attr-defined]
        self.assertEqual(items[2].tags_required, [])  # type: ignore[attr-defined]
        self.assertIsNone(items[2].subject)  # type: ignore[attr-defined]
        self.assertEqual(items[2].context, "context")  # type: ignore[attr-defined]

        self.assertEqual(items[3].tags_provided, ["a", "b"])  # type: ignore[attr-defined]
        self.assertEqual(items[3].tags_required, [])  # type: ignore[attr-defined]
        self.assertEqual(items[3].subject, "provided")  # type: ignore[attr-defined]
        self.assertIsNone(items[3].context)  # type: ignore[attr-defined]

        self.assertEqual(items[4].tags_provided, [])  # type: ignore[attr-defined]
        self.assertEqual(items[4].tags_required, ["c", "d"])  # type: ignore[attr-defined]
        self.assertEqual(items[4].subject, "required")  # type: ignore[attr-defined]
        self.assertIsNone(items[4].context)  # type: ignore[attr-defined]

        self.assertEqual(items[5].tags_provided, [])  # type: ignore[attr-defined]
        self.assertEqual(items[5].tags_required, [])  # type: ignore[attr-defined]
        self.assertEqual(items[5].subject, "exprs")  # type: ignore[attr-defined]
        self.assertIsNone(items[5].context)  # type: ignore[attr-defined]

        self.assertEqual(len(items), 6)

    def test_task_type_name_filter(self) -> None:
        noop = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop"
            )
        )
        self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="other"
            )
        )
        self.assertEqual([x.id for x in self.lookup().candidates], [noop.id])

    def test_subject_context_filter(self) -> None:
        none = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop"
            )
        )
        subj = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop", subject="subject"
            )
        )
        cont = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop", context="context"
            )
        )
        both = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="subject",
                context="context",
            )
        )

        for subject, context, expected in (
            (None, None, [none]),
            ("subject", None, [none, subj]),
            (None, "context", [none, cont]),
            ("subject", "context", [none, cont, subj, both]),
        ):
            with self.subTest(subject=subject, context=context):
                self.assertQuerySetEqual(
                    self.lookup(subject=subject, context=context).candidates,
                    expected,
                    ordered=False,
                )

    def test_tags_filter(self) -> None:
        none = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="none",
            )
        )
        prov = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="prov",
                tags_provided=["p"],
            )
        )
        req = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="req",
                tags_required=["r"],
            )
        )
        both = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="both",
                tags_provided=["p"],
                tags_required=["r"],
            )
        )
        self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="enotmatch",
                tags_provided="a or b",
                tags_required="c or d",
            )
        )
        eprov = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="eprov",
                tags_provided="p and not q",
            )
        )
        ereq = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="ereq",
                tags_required="r and not q",
            )
        )
        eboth = self._add_config(
            DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="eboth",
                tags_provided="p and not q",
                tags_required="r and not q",
            )
        )

        for provided, required, expected in (
            ("", "", [none]),
            ("p", "", [none, prov, eprov]),
            ("pqr", "", [none, prov]),
            ("", "r", [none, req, ereq]),
            ("", "pqr", [none, req]),
            ("p", "r", [none, prov, req, both, eprov, ereq, eboth]),
            ("pqr", "pqr", [none, prov, req, both]),
        ):
            with (
                self.subTest(provided=provided, required=required),
                mock.patch(
                    "debusine.server.collections.debusine_task_configuration.ConfigurationLookup.set_subject_context_constraints"
                ),
            ):
                self.assertEqual(
                    [
                        x.subject
                        for x in self.lookup(
                            tags_provided=set(provided),
                            tags_required=set(required),
                        ).iter_matching_entries()
                    ],
                    [x.data["subject"] for x in expected],
                )

    def test_candidates_deleted_entries(self) -> None:
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.manager.remove_item(
            self._add_config(entry), user=self.scenario.user
        )
        active = self._add_config(entry)
        self.assertQuerySetEqual(self.lookup().candidates, [active])

    def test_iter_matching_entries_ordering_path(self) -> None:
        self.enterContext(
            mock.patch(
                "debusine.server.collections.debusine_task_configuration.ConfigurationLookup.set_subject_context_constraints"
            )
        )
        self._add_config(
            entry1 := DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop", path="03.yaml"
            )
        )
        self._add_config(
            entry2 := DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                path="02.yaml",
                subject="subject",
            )
        )
        self._add_config(
            entry3 := DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop", context="context"
            )
        )
        assert list(self.lookup().iter_matching_entries()) == [
            entry3,
            entry2,
            entry1,
        ]

    def test_iter_matching_entries_ordering_position(self) -> None:
        self.enterContext(
            mock.patch(
                "debusine.server.collections.debusine_task_configuration.ConfigurationLookup.set_subject_context_constraints"
            )
        )
        self._add_config(
            entry1 := DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                path="f.yaml",
                position=3,
            )
        )
        self._add_config(
            entry2 := DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                path="f.yaml",
                position=2,
                subject="subject",
            )
        )
        self._add_config(
            entry3 := DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                context="context",
                path="f.yaml",
            )
        )
        assert list(self.lookup().iter_matching_entries()) == [
            entry3,
            entry2,
            entry1,
        ]

    def test_iter_matching_entries_ordering_legacy(self) -> None:
        self.enterContext(
            mock.patch(
                "debusine.server.collections.debusine_task_configuration.ConfigurationLookup.set_subject_context_constraints"
            )
        )
        self._add_config(
            entry1 := DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER, task_name="noop"
            )
        )
        self._add_config(
            entry2 := DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="subject",
            )
        )
        self._add_config(
            entry3 := DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                context="context",
            )
        )
        self._add_config(
            entry4 := DebusineTaskConfiguration(
                task_type=TaskTypes.WORKER,
                task_name="noop",
                subject="subject",
                context="context",
            )
        )
        assert list(self.lookup().iter_matching_entries()) == [
            entry1,
            entry3,
            entry2,
            entry4,
        ]

    def test_iter_matching_entries_corrupted(self) -> None:
        """Lookup raises if the item contains invalid data."""
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        item = self._add_config(entry)
        item.data["invalid_field"] = "value"
        item.save()
        with self.assertRaisesRegex(
            TaskConfigError,
            rf"(?s)debusine:task-configuration item {item.id} does not validate"
            r".+Extra inputs are not permitted",
        ):
            list(self.lookup().iter_matching_entries())

    def test_lookup_templates_empty(self) -> None:
        """Lookup templates with use_templates is empty."""
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self.assertEqual(
            list(self.lookup().lookup_templates(entry)),
            [],
        )

    def test_lookup_templates_missing_item(self) -> None:
        """Lookup templates with use_templates is empty."""
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            use_templates=["missing"],
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            r"Worker:noop:: references missing template 'missing'",
        ):
            list(self.lookup().lookup_templates(entry))

    def test_lookup_templates_found(self) -> None:
        """Lookup templates with one template."""
        template = DebusineTaskConfiguration(
            template="template",
        )
        self._add_config(template)

        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            use_templates=["template"],
        )
        self.assertEqual(
            list(self.lookup().lookup_templates(entry)),
            [template],
        )

    def test_lookup_templates_corrupted(self) -> None:
        template = DebusineTaskConfiguration(
            template="template",
        )
        item = self._add_config(template)
        item.data["corrupted"] = True
        item.save()

        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            use_templates=["template"],
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            r"(?s)debusine:task-configuration item does not validate"
            r".+Extra inputs are not permitted",
        ):
            list(self.lookup().lookup_templates(entry))

    def test_lookup_templates_diamond(self) -> None:
        """Lookup templates with a diamond dependency."""
        base = DebusineTaskConfiguration(
            template="base",
        )
        self._add_config(base)
        a = DebusineTaskConfiguration(template="a", use_templates=["base"])
        self._add_config(a)
        b = DebusineTaskConfiguration(template="b", use_templates=["base"])
        self._add_config(b)
        ab = DebusineTaskConfiguration(template="ab", use_templates=["a", "b"])
        self._add_config(ab)

        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            use_templates=["ab"],
        )
        self.assertEqual(
            list(self.lookup().lookup_templates(entry)),
            [base, a, base, b, ab],
        )

    def test_lookup_templates_cycle(self) -> None:
        """Lookup templates with a dependency cycle."""
        a = DebusineTaskConfiguration(template="a", use_templates=["b"])
        self._add_config(a)
        b = DebusineTaskConfiguration(template="b", use_templates=["a"])
        self._add_config(b)

        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            use_templates=["a"],
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            r"template:b: template lookup cycle detected: a→b→a",
        ):
            list(self.lookup().lookup_templates(entry))

    def test_iter_items_no_templates(self) -> None:
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        self._add_config(entry)
        self.assertEqual(
            list(self.lookup().iter_items()),
            [entry],
        )

    def test_iter_items_empty(self) -> None:
        self.assertEqual(list(self.lookup().iter_items()), [])

    def test_iter_items_templates(self) -> None:
        a = DebusineTaskConfiguration(template="a")
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            use_templates=["a"],
        )
        self._add_config(entry)
        with mock.patch(
            "debusine.server.collections"
            ".debusine_task_configuration.ConfigurationLookup.lookup_templates",
            return_value=iter((a, a, a)),
        ):
            self.assertEqual(
                list(self.lookup().iter_items()),
                [entry, a, a, a],
            )


class BuildConfigurationTests(unittest.TestCase):
    """Test :py:func:build_configuration`."""

    def entry(
        self, name: str = "test", **kwargs: Any
    ) -> DebusineTaskConfiguration:
        return DebusineTaskConfiguration(template=name, **kwargs)

    def test_empty(self) -> None:
        """Run with no entries."""
        self.assertEqual(
            build_configuration(()),
            Changes(
                default_values={},
                override_values={},
                provide_tags=set(),
                require_tags=set(),
            ),
        )

    def test_one(self) -> None:
        """Run with one entry."""
        entry = self.entry(
            delete_values=["val0"],
            default_values={"val1": 1},
            override_values={"val2": 2},
            lock_values=["val3"],
        )
        self.assertEqual(
            build_configuration([entry]),
            Changes(
                default_values={"val1": 1},
                override_values={"val2": 2},
                provide_tags=set(),
                require_tags=set(),
            ),
        )

    def test_default_and_override(self) -> None:
        """Run with one entry with same default and override keys."""
        entry = self.entry(
            default_values={"val1": 1},
            override_values={"val1": 11},
        )
        self.assertEqual(
            build_configuration([entry]),
            Changes(
                default_values={"val1": 1},
                override_values={"val1": 11},
                provide_tags=set(),
                require_tags=set(),
            ),
        )

    def test_stack(self) -> None:
        """Run with two entries."""
        entry1 = self.entry(
            default_values={"val1": 1, "base1": 1},
            override_values={"val1": 2, "base2": 2},
        )
        entry2 = self.entry(
            default_values={"val1": 11},
            override_values={"val1": 22},
        )
        self.assertEqual(
            build_configuration([entry1, entry2]),
            Changes(
                default_values={"val1": 11, "base1": 1},
                override_values={"val1": 22, "base2": 2},
                provide_tags=set(),
                require_tags=set(),
            ),
        )

    def test_delete_values(self) -> None:
        """Test delete_values."""
        entry1 = self.entry(
            default_values={"val1": 1, "val2": 2},
            override_values={"val1": 11, "val2": 22},
        )
        entry2 = self.entry(
            delete_values=["val1"],
        )
        self.assertEqual(
            build_configuration([entry1, entry2]),
            Changes(
                default_values={"val2": 2},
                override_values={"val2": 22},
                provide_tags=set(),
                require_tags=set(),
            ),
        )

    def test_delete_and_set_values(self) -> None:
        """Test deleting and then setting values."""
        entry1 = self.entry(
            default_values={"val1": 1, "val2": 2},
            override_values={"val1": 11, "val2": 22},
        )
        entry2 = self.entry(
            delete_values=["val1"],
            default_values={"val1": 111},
            override_values={"val1": 112},
        )
        self.assertEqual(
            build_configuration([entry1, entry2]),
            Changes(
                default_values={"val1": 111, "val2": 2},
                override_values={"val1": 112, "val2": 22},
                provide_tags=set(),
                require_tags=set(),
            ),
        )

    def test_lock_values(self) -> None:
        """Test locked_values."""
        entry1 = self.entry(
            default_values={"val1": 1, "val2": 2},
            override_values={"val1": 11, "val2": 22},
            lock_values=["val1"],
        )
        entry2 = self.entry(
            delete_values=["val1", "val2"],
            default_values={"val1": 111, "val3": 3},
            override_values={"val1": 112, "val3": 33},
        )
        self.assertEqual(
            build_configuration([entry1, entry2]),
            Changes(
                default_values={"val1": 1, "val3": 3},
                override_values={"val1": 11, "val3": 33},
                provide_tags=set(),
                require_tags=set(),
            ),
        )

    def test_tags(self) -> None:
        entry1 = self.entry(
            provide_tags=["pa", "pb"], require_tags=["ra", "rb"]
        )
        entry2 = self.entry(
            provide_tags=["pa", "pc"], require_tags=["ra", "rc"]
        )
        self.assertEqual(
            build_configuration([entry1, entry2]),
            Changes(
                default_values={},
                override_values={},
                provide_tags={"pa", "pb", "pc"},
                require_tags={"ra", "rb", "rc"},
            ),
        )


class ApplyConfigurationTests(TaskConfigTestCase):
    """Test :py:func:apply_configuration`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task = mock.MagicMock(spec=DBTask)
        self.task.TASK_TYPE = TaskTypes.WORKER
        self.task.name = "noop"
        self.task.get_task_configuration_subject_context.return_value = (
            None,
            None,
        )
        self.config_tags_provided = ttags.TaskProvidedTags()
        self.config_tags_required = wtags.TaskRequiredTags()
        self.task.compute_scheduler_tags.return_value = (
            self.config_tags_provided,
            self.config_tags_required,
        )

    def test_no_config(self) -> None:
        """Test without configuration to apply."""
        task_data: dict[str, Any] = {}
        apply_configuration(task_data, self.collection, self.task)
        self.assertEqual(task_data, {"task_configuration": self.collection.pk})

    def test_with_config(self) -> None:
        """Test with overrides to apply."""
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            default_values={"default": "added"},
            override_values={"override": "overridden"},
        )
        self._add_config(entry)

        task_data: dict[str, Any] = {
            "key": "value",
            "override": "original",
        }
        apply_configuration(task_data, self.collection, self.task)
        self.assertEqual(
            task_data,
            {
                "key": "value",
                "override": "overridden",
                "default": "added",
                "task_configuration": self.collection.pk,
            },
        )

    def test_defaults_with_none(self) -> None:
        """Test setting defaults on existing values set to None."""
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            default_values={"default": "added"},
        )
        self._add_config(entry)

        task_data: dict[str, Any] = {"default": None}
        apply_configuration(task_data, self.collection, self.task)
        self.assertEqual(
            task_data,
            {"default": "added", "task_configuration": self.collection.pk},
        )

    def test_defaults_preserves_existing_value(self) -> None:
        entry = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            default_values={"default": "added"},
        )
        self._add_config(entry)

        task_data: dict[str, Any] = {"default": "already set"}
        apply_configuration(task_data, self.collection, self.task)
        self.assertEqual(
            task_data,
            {
                "default": "already set",
                "task_configuration": self.collection.pk,
            },
        )

    def test_scheduler_tags_passed_to_lookup(self) -> None:
        for tags_provided, tags_required in (
            ([], []),
            (["a"], []),
            ([], ["a"]),
            (["a"], ["b"]),
        ):
            with (
                self.subTest(provided=tags_provided, required=tags_required),
                mock.patch(
                    "debusine.server.collections.debusine_task_configuration"
                    ".ConfigurationLookup.set_tags_provided"
                ) as set_tags_provided,
                mock.patch(
                    "debusine.server.collections.debusine_task_configuration"
                    ".ConfigurationLookup.set_tags_required"
                ) as set_tags_required,
            ):
                self.config_tags_provided.sets = {
                    ttags.ProvenanceProvided.SYSTEM: set(tags_provided),
                }
                self.config_tags_required.sets = {
                    wtags.ProvenanceRequired.SYSTEM: set(tags_required),
                }
                apply_configuration({}, self.collection, self.task)
                set_tags_provided.assert_called_once_with(set(tags_provided))
                set_tags_required.assert_called_once_with(set(tags_required))

    def test_issue1423(self) -> None:
        source = self.playground.create_minimal_source_package_artifact()
        environment = self.playground.create_debian_environment(
            codename="bookworm"
        )
        assert environment.artifact is not None
        bookworm = environment.artifact
        coll = self.playground.create_collection(
            "overlay-not-automatic", CollectionCategory.TASK_CONFIGURATION
        )
        coll.manager.add_bare_data(
            BareDataCategory.TASK_CONFIGURATION,
            user=self.scenario.user,
            data=DebusineTaskConfiguration(
                override_values={"build_dep_resolver": "aptitude"},
                task_name="sbuild",
                task_type=TaskTypes.WORKER,
            ),
        )

        wr = self.playground.create_worker_task(
            task_name="sbuild",
            task_data=SbuildData(
                input=SbuildInput(source_artifact=source.pk),
                backend="unshare",
                environment=bookworm.pk,
                build_components=["any"],
                build_architecture="amd64",
                extra_repositories=[
                    {
                        "url": "http://deb.debian.org/debian",
                        "suite": "bookworm",
                        "components": [
                            "contrib",
                            "non-free",
                            "non-free-firmware",
                        ],
                    },
                    {
                        "url": "http://deb.debian.org/debian",
                        "suite": "bookworm-backports",
                        "components": [
                            "main",
                            "contrib",
                            "non-free",
                            "non-free-firmware",
                        ],
                    },
                    {
                        "url": "http://incoming.debian.org/debian-buildd",
                        "suite": "buildd-bookworm-backports",
                        "components": [
                            "main",
                            "contrib",
                            "non-free",
                            "non-free-firmware",
                        ],
                    },
                ],
                task_configuration="overlay-not-automatic",
            ),
        )
        self.assertIsNone(wr.configured_task_data)
        self.playground.advance_work_request(wr, mark_pending=True)
        assert wr.configured_task_data is not None
        self.assertNotIn("build_dep_resolver", wr.task_data)
        self.assertEqual(
            wr.configured_task_data["build_dep_resolver"], "aptitude"
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""The unshare executor backend."""

import os
import shlex
import shutil
import subprocess
import tempfile
from pathlib import Path, PurePath
from subprocess import CompletedProcess
from typing import Any, AnyStr, Literal, overload, override

import psutil

from debusine import utils
from debusine.client.debusine import Debusine
from debusine.client.models import ArtifactResponse
from debusine.tasks.executors.base import (
    ExecutorImageCategory,
    ExecutorInterface,
    ExecutorStatistics,
    ImageNotDownloadedError,
    InstanceInterface,
    InstanceNotRunning,
)
from debusine.tasks.executors.images import ImageCache
from debusine.tasks.models import BackendType


class UnshareExecutor(ExecutorInterface, backend_name=BackendType.UNSHARE):
    """Support the unshare(1) executor."""

    system_image: ArtifactResponse
    working_directory: Path
    _image_cache: ImageCache
    _local_path: Path | None = None

    image_category = ExecutorImageCategory.TARBALL

    def __init__(
        self,
        debusine_api: Debusine,
        system_image_id: int,
        workspace: str,
        working_directory: Path,
    ) -> None:
        """
        Instantiate an UnshareExecutor.

        :param debusine_api: The object to use the debusine client API.
        :param system_image_id: An artifact id pointing to the system tarball.
        """
        self._image_cache = ImageCache(
            debusine_api, self.image_category, workspace=workspace
        )
        self._extracted: Path | None = None
        self.system_image = self._image_cache.image_artifact(system_image_id)
        self.working_directory = working_directory

    @override
    @classmethod
    def available(cls) -> bool:
        return all(
            utils.is_command_available(cmd)
            for cmd in (UnshareInstance.unshare_command, "ip")
        )

    @override
    def download_image(self) -> str:
        self._local_path = self._image_cache.download_image(self.system_image)
        return self.image_name()

    @override
    def image_name(self) -> str:
        """Return the path to the downloaded image."""
        if self._local_path is None:
            raise ImageNotDownloadedError()
        return str(self._local_path)

    @override
    def create(self) -> "UnshareInstance":
        """
        Create an UnshareInstance using the downloaded image.

        Returns a new, stopped UnshareInstance.
        """
        if self._local_path is None:
            raise ImageNotDownloadedError(
                "UnshareExecutor.download_image() should have "
                "been called and set self._local_path"
            )

        return UnshareInstance(self._local_path, self.working_directory)

    @override
    def autopkgtest_virt_server(self) -> str:
        return "unshare"

    @override
    def autopkgtest_virt_args(self) -> list[str]:
        if self._local_path is None:
            raise ImageNotDownloadedError()
        return [
            "--arch",
            self.system_image.data["architecture"],
            "--release",
            self.system_image.data["codename"],
            "--tarball",
            str(self._local_path),
        ]

    @override
    @classmethod
    def get_statistics(
        cls, instance: InstanceInterface | None
    ) -> ExecutorStatistics:
        """Return statistics for this instance."""
        statistics = ExecutorStatistics()

        if instance is not None and instance.is_started():
            assert isinstance(instance, UnshareInstance)
            try:
                statvfs = os.statvfs(instance._root)
            except FileNotFoundError:
                pass
            else:
                statistics.disk_space = (
                    statvfs.f_blocks - statvfs.f_bfree
                ) * statvfs.f_frsize
                statistics.available_disk_space = (
                    statvfs.f_blocks * statvfs.f_frsize
                )

        memory = psutil.virtual_memory()
        statistics.memory = memory.used
        statistics.available_memory = memory.total

        statistics.cpu_count = psutil.cpu_count() or 1

        return statistics


class UnshareInstance(InstanceInterface):
    """Support instances of the unshare(1) executor."""

    unshare_command = "unshare"
    _unshare_default_args = ["--map-users=auto", "--map-groups=auto"]

    def __init__(self, image: Path, working_directory: Path) -> None:
        """Initialize the object."""
        super().__init__()
        self._image = image
        self._extracted: Path | None = None
        self.working_directory: Path = working_directory

    @property
    def _root(self) -> Path:
        if self._extracted is None:
            raise InstanceNotRunning(
                "UnshareInstance.start() must be called before using _root"
            )
        return self._extracted / "root"

    @override
    def is_started(self) -> bool:
        return self._extracted is not None

    @override
    def do_start(self) -> None:
        """Start the instance, making it ready to run commands."""
        # extracted directory is deleted by UnshareInstance.stop()
        self._extracted = Path(
            tempfile.mkdtemp(prefix="unshare-", dir=self.working_directory)
        )

        # The directory, by default, has permissions 700. To extract the
        # file using subuids it needs wider permissions
        # If the directory that mkdtemp() used is restricted to the invoking
        # user (private tmp): then the permission of self._extracted.parent
        # may be insufficient to allow extraction in that namespace.
        self._extracted.chmod(0o755)

        self._root.mkdir()

        # Make / dir of the environment owned by "root:root" (1:1) in the
        # environment namespace
        # E.g. debusine-worker owns self._root. After the chown it's owned
        # by the first subuid:subgid of the user (which will be root:root
        # in the chroot)
        self._run_unshare_command(
            ["--map-root-user", "chown", "1:1", str(self._root)]
        )

        with self._image.open("rb") as image:
            self._run_unshare_command(
                [
                    "--setuid=0",
                    "--setgid=0",
                    "tar",
                    "-C",
                    str(self._root),
                    "--xz",
                    "-x",
                ],
                stdin=image,
            )
        # Copy the system resolv.conf into the instance, overriding any
        # systemd-resolved symlink
        self._run_unshare_command(
            [
                "--setuid=0",
                "--setgid=0",
                "rm",
                "--force",
                str(self._root / "etc" / "resolv.conf"),
            ]
        )
        self._run_unshare_command(
            [
                "--setuid=0",
                "--setgid=0",
                "cp",
                "--dereference",
                "/etc/resolv.conf",
                str(self._root / "etc" / "resolv.conf"),
            ]
        )

    @override
    def do_stop(self) -> None:
        """Stop the instance: delete the directory."""
        assert self._extracted is not None
        self._run_unshare_command(
            ["--map-root-user", "rm", "-rf", str(self._root)]
        )
        shutil.rmtree(self._extracted)

        self._extracted = None

    @override
    def do_restart(self) -> None:
        """Restart the environment (noop for unshare)."""

    @override
    def do_file_push(
        self, source: Path, target: PurePath, uid: int, gid: int, mode: int
    ) -> None:
        inner_target = shlex.quote(str(self._root / target.relative_to("/")))

        with source.open("rb") as stdin:
            self._run_unshare_command(
                [
                    f"--setuid={uid}",
                    f"--setgid={gid}",
                    "sh",
                    "-c",
                    f"cat > {inner_target} && chmod {mode:o} {inner_target}",
                ],
                stdin=stdin,
                check=True,
            )

    @override
    def do_file_pull(self, source: PurePath, target: Path) -> None:
        self._run_unshare_command(
            [
                "--setuid=0",
                "--setgid=0",
                "--map-root-user",
                "cp",
                "--no-preserve=ownership,mode",
                str(self._root / source.relative_to("/")),
                str(target),
            ],
            check=True,
        )

    @overload
    def do_run(
        self,
        args: list[str],
        text: Literal[True],
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[str]: ...

    @overload
    def do_run(
        self,
        args: list[str],
        text: Literal[False] | None = None,
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[bytes]: ...

    @override
    def do_run(
        self,
        args: list[str],
        text: Literal[False] | Literal[True] | None = None,
        cwd: str | os.PathLike[str] | None = None,
        run_as_root: bool = False,
        **kwargs: Any,
    ) -> CompletedProcess[AnyStr]:
        if cwd is not None:
            # TODO: Once we can require environments to have coreutils >=
            # 2.28 (Debian 10 or newer), we can use `env --chdir` instead.
            args = [
                "sh",
                "-c",
                f"cd {shlex.quote(str(cwd))} && {shlex.join(args)}",
            ]

        if not run_as_root:
            username = self.non_root_user
            self.create_user()
            args = ["/sbin/runuser", "-u", username, "--", *args]

        root = shlex.quote(str(self._root))
        args_quoted = shlex.join(args)
        shell_cmd = (
            f"mount --rbind /dev {root}/dev && "
            f"exec /usr/sbin/chroot {root} {args_quoted}"
        )

        kwargs_with_text: dict[str, Any] = {**kwargs, "text": text}

        return self._run_unshare_command(
            [
                "--setuid=0",
                "--setgid=0",
                "--kill-child",
                "--pid",
                "--mount",
                f"--mount-proc={root}/proc",
                "sh",
                "-c",
                shell_cmd,
            ],
            **kwargs_with_text,
        )

    @overload
    @classmethod
    def _run_unshare_command(
        cls,
        args: list[str],
        *,
        text: Literal[True],
        **kwargs: Any,
    ) -> CompletedProcess[str]: ...

    @overload
    @classmethod
    def _run_unshare_command(
        cls,
        args: list[str],
        *,
        text: Literal[False] | None = None,
        **kwargs: Any,
    ) -> CompletedProcess[bytes]: ...

    @classmethod
    def _run_unshare_command(
        cls, args: list[str], **kwargs: Any
    ) -> CompletedProcess[AnyStr]:
        """
        Run args in unshare.

        kwargs are passed to subprocess.run. By default, "stdout" and "stderr"
        are subprocess.PIPE.

        Uses cls.unshare_command and cls._unshare_default_args for the command
        and default arguments.
        """
        cmd = [cls.unshare_command, *cls._unshare_default_args, *args]

        extra_kwargs: dict[str, Any] = {}

        if "stderr" not in kwargs:
            extra_kwargs["stderr"] = subprocess.PIPE

        if "stdout" not in kwargs:
            extra_kwargs["stdout"] = subprocess.PIPE

        if kwargs.get("env") is None:
            # Run commands in a pristine environment by default
            kwargs["env"] = {}

        # Include split-/usr paths to support older releases (< bookworm)
        kwargs["env"]["PATH"] = (
            "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
        )

        return subprocess.run(cmd, **kwargs, **extra_kwargs)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the Worker."""

import asyncio
import functools
import logging
import re
import secrets
import signal
import socket
import subprocess
from collections.abc import AsyncGenerator, Coroutine
from pathlib import Path
from textwrap import dedent
from typing import Any, AnyStr, Literal, Never, override
from unittest import mock
from unittest.mock import MagicMock

import aiohttp
from aiohttp import RequestInfo
from aiohttp.client_reqrep import ConnectionKey
from aiohttp.web import Response
from aiohttp.web_exceptions import (
    HTTPForbidden,
    HTTPInternalServerError,
    HTTPServiceUnavailable,
)
from multidict import CIMultiDict, CIMultiDictProxy
from yarl import URL

import debusine.tasks.models as task_models
from debusine.artifacts.models import WorkRequestResults
from debusine.client.server_notifications import ServerNotifications
from debusine.tasks import Sbuild, TaskConfigError
from debusine.tasks.models import OutputData, OutputDataError, WorkerType
from debusine.test import TestCase
from debusine.worker import Worker, _worker
from debusine.worker.config import ConfigHandler
from debusine.worker.system_information import system_metadata
from debusine.worker.task_runner import TaskRunner
from debusine.worker.tests import server
from debusine.worker.tests.server import ServerConfig


class WorkerConfigTestCase(TestCase, server.DebusineAioHTTPTestCase):
    """Test case with a worker configuration."""

    @override
    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.api_url = str(self.server.make_url('')) + "/api"

        self.temp_config_directory = self.create_temp_config_directory(
            {'General': {'api-url': self.api_url}}
        )

        self.config = ConfigHandler(
            directories=[self.temp_config_directory.as_posix()]
        )


class WorkerTestCase(WorkerConfigTestCase):
    """Base test case for Worker tests."""

    @override
    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.default_sigint_handler = signal.getsignal(signal.SIGINT)
        self.default_sigterm_handler = signal.getsignal(signal.SIGTERM)

    @override
    def tearDown(self) -> None:
        # Restore signal handlers. Cli.execute() changes them
        signal.signal(signal.SIGINT, self.default_sigint_handler)
        signal.signal(signal.SIGTERM, self.default_sigterm_handler)
        super().tearDown()


class WorkerRegistrationTests(WorkerTestCase):
    """Test worker registration."""

    worker: Worker

    @override
    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.worker = await self.enterAsyncContext(
            Worker(
                worker_type=WorkerType.EXTERNAL,
                working_directory=self.temp_config_directory.as_posix(),
                config=self.config,
            )
        )

        # Don't risk triggering local hooks
        self.worker.AFTER_WORK_REQUEST_HOOK_DIR = Path("/nonexistent")

    def patch_http_method(self, http_method: str, **kwargs: Any) -> MagicMock:
        """Patch a specific HTTP method of Worker's HTTP Client."""
        # setup_worker must be called first.
        assert self.worker is not None

        mocked = self.enterContext(
            mock.patch.object(
                self.worker._async_http_client, http_method, autospec=True
            )
        )
        for key, value in kwargs.items():
            setattr(mocked, key, value)
        return mocked

    async def test_registration_success(self) -> None:
        """
        Worker client registers successfully.

        Client creates a token, sends it to the server and save to the
        configuration.
        """
        # No token in the configuration
        self.assertIsNone(self.config.token)

        # Worker connects: it will create a token, send and save
        await self.worker.require_token()

        # A token has been retrieved
        self.assertIsNotNone(self.config.token)

        registration_request = self.server_requests[-1]
        self.assertEqual(registration_request.path, '/api/1.0/worker/register/')

        # Server side received expected registration data
        self.assertEqual(
            registration_request.json_content,
            {
                'token': self.config.token,
                'fqdn': socket.getfqdn(),
                'worker_type': WorkerType.EXTERNAL,
            },
        )
        assert registration_request.headers is not None
        self.assertNotIn('token', registration_request.headers)

    async def test_registration_success_activation_token(self) -> None:
        """Worker client registers successfully with an activation token."""
        activation_token = secrets.token_hex(32)
        activation_token_path = Path(self.config._activation_token_file)
        activation_token_path.write_bytes(activation_token.encode())  # noqa: ASYNC240
        activation_token_path.chmod(0o600)  # noqa: ASYNC240
        self.assertEqual(self.config.activation_token, activation_token)

        # No token in the configuration
        self.assertIsNone(self.config.token)

        # Worker connects: it will create a token, send and save
        await self.worker.require_token()

        # A token has been retrieved
        self.assertIsNotNone(self.config.token)

        registration_request = self.server_requests[-1]
        self.assertEqual(registration_request.path, '/api/1.0/worker/register/')

        # Server side received expected registration data
        self.assertEqual(
            registration_request.json_content,
            {
                'token': self.config.token,
                'fqdn': socket.getfqdn(),
                'worker_type': WorkerType.EXTERNAL,
            },
        )
        assert registration_request.headers is not None
        self.assertEqual(
            registration_request.headers['token'], activation_token
        )

    async def test_registration_failure(self) -> None:
        """
        Worker client registration failure.

        The server returns 403, the client raises WorkerRegistrationError.
        """
        self.server_config.register_response_status_code = (
            HTTPForbidden.status_code
        )

        log_message = (
            'Could not register. Server HTTP status code: 403 Body: error'
        )

        with (
            self.assertLogsContains(log_message, logger="debusine.worker"),
            self.assertRaisesSystemExit(3),
        ):
            await self.worker.require_token()

    async def test_client_connector_error(self) -> None:
        """Worker cannot connect to the server: server unreachable."""
        connection_key = mock.create_autospec(spec=ConnectionKey)
        mock_post = self.patch_http_method(
            "post",
            side_effect=aiohttp.client_exceptions.ClientConnectorError(
                connection_key, OSError("Unreachable")
            ),
        )
        log_message = "Could not register. Server unreachable: "
        with (
            self.assertLogsContains(log_message, logger="debusine.worker"),
            self.assertRaisesSystemExit(3),
        ):
            await self.worker.require_token()

        self.assertTrue(mock_post.called)

    async def test_connection_read_error(self) -> None:
        """Worker cannot read data from the server (low level disconnection)."""
        mock_post = self.patch_http_method(
            "post",
            side_effect=aiohttp.client_exceptions.ClientResponseError(
                RequestInfo(
                    URL('https://test.com'),
                    '',
                    CIMultiDictProxy(CIMultiDict()),
                    URL('https://test.com'),
                ),
                (),
            ),
        )
        log_message = (
            'Could not register. Server HTTP status code: 0 Body: Not available'
        )
        with (
            self.assertLogsContains(log_message, logger="debusine.worker"),
            self.assertRaisesSystemExit(3),
        ):
            await self.worker.require_token()

        self.assertTrue(mock_post.called)

    async def test_connection_rejected(self) -> None:
        """Worker tries to connect, server returns 500."""
        self.server_config.register_response_status_code = (
            HTTPInternalServerError.status_code
        )

        log_message = (
            'Could not register. Server HTTP status code: 500 Body: error'
        )

        with (
            self.assertLogsContains(log_message, logger="debusine.worker"),
            self.assertRaisesSystemExit(3),
        ):
            await self.worker.require_token()


class WorkerEnvTests(WorkerTestCase):
    """Test Worker environment setup/teardown."""

    worker: Worker

    async def setup_worker(
        self,
        *,
        log_level: str | None = None,
        working_directory: Path | None = None,
        setup_logging: bool = True,
        use_working_directory: bool = True,
    ) -> None:
        kwargs = {}
        if use_working_directory:
            kwargs["working_directory"] = (
                working_directory or self.temp_config_directory
            ).as_posix()

        if not setup_logging:
            # Disable the logging configuration, it makes it hard to test the
            # logs
            self.enterContext(mock.patch.object(Worker, "_setup_logging"))

        self.worker = await self.enterAsyncContext(
            Worker(
                log_level=log_level,
                worker_type=WorkerType.EXTERNAL,
                config=self.config,
                **kwargs,
            )
        )

        # Don't risk triggering local hooks
        self.worker.AFTER_WORK_REQUEST_HOOK_DIR = Path("/nonexistent")

    def patch_config_handler(self) -> MagicMock:
        """
        Patch ConfigHandler class.

        :return: the MagicMock
        """
        config_handler_mocked = self.enterContext(
            mock.patch("debusine.worker._worker.ConfigHandler", autospec=True)
        )
        instance = config_handler_mocked.return_value
        instance.working_directory = Path('/tmp')
        return config_handler_mocked

    def assertQueuedEvents(self, expected: list[str]) -> None:
        """
        Check that the queued events match the given list.

        Events are stringified before comparison
        """
        actual: list[str] = []
        while not self.worker.event_queue.empty():
            evt = self.worker.event_queue.get_nowait()
            assert isinstance(evt, _worker.Event)
            actual.append(str(evt))
        assert actual == expected

    def test_worker_init(self) -> None:
        """Init without any keyword."""
        mocked_config_handler = self.patch_config_handler()

        worker = Worker(
            log_level="WARNING",
            working_directory=self.temp_config_directory.as_posix(),
        )

        # Worker() created a ConfigHandler default object and set it
        # to Worker._config
        self.assertIsNotNone(worker._config)
        mocked_config_handler.assert_called_once_with()

    def test_worker_init_signing(self) -> None:
        """Signing workers use different dirs and require an HTTPS api-url."""
        mocked_config_handler = self.patch_config_handler()

        worker = Worker(
            log_level="WARNING",
            worker_type=WorkerType.SIGNING,
            working_directory=self.temp_config_directory.as_posix(),
        )

        self.assertIsNotNone(worker._config)
        mocked_config_handler.assert_called_once_with(
            directories=[
                str(Path.home() / ".config/debusine/signing"),
                "/etc/debusine/signing",
            ],
            require_https=True,
        )

    async def test_worker_init_use_config_handler(self) -> None:
        """worker.set_config_handler() sets the ConfigHandler."""
        await self.setup_worker(log_level="WARNING")
        assert self.worker is not None

        self.assertIs(self.worker._config, self.config)

    async def test_worker_init_bad_working_directory(self) -> None:
        with (
            self.assertLogsContains(
                "could not create working directory", logger="debusine.worker"
            ),
            self.assertRaisesSystemExit(3),
        ):
            await self.setup_worker(
                working_directory=Path("/nonexistent"),
                setup_logging=False,
            )

    async def test_worker_init_working_directory_from_config(self) -> None:
        mydir = self.temp_config_directory / "foobar"
        self.config['General']['working-directory'] = str(mydir)

        await self.setup_worker(use_working_directory=False)

        assert self.worker is not None
        self.assertEqual(self.worker._workdir, mydir)

    async def test_clean_working_directory(self) -> None:
        """worker._clean_workdir() cleans up working directory."""
        await self.setup_worker()
        assert self.worker is not None
        # Files that will be dropped
        work_request_dir = self.temp_config_directory / "WR-123"
        sample_file = work_request_dir / "sample.txt"
        work_request_dir.mkdir(parents=True)
        sample_file.write_text("sample text")
        # Files that will be kept
        random_dir = self.temp_config_directory / "random"
        random_dir.mkdir()

        self.worker.clean_workdir()

        self.assertFalse(sample_file.exists())
        self.assertFalse(work_request_dir.exists())
        self.assertTrue(random_dir.exists())

    async def test_clean_working_directory_with_failing_rmtree(self) -> None:
        """worker._clean_workdir() fails gracefully when rmtree fails."""
        await self.setup_worker()
        assert self.worker is not None
        # Files that will be dropped
        work_request_dir = self.temp_config_directory / "WR-123"
        work_request_dir.mkdir(parents=True)
        with (
            self.assertLogsContains(
                f"Failed to remove '{work_request_dir}'",
                logger="debusine.worker",
            ),
            mock.patch("shutil.rmtree") as mock_rmtree,
        ):
            mock_rmtree.side_effect = OSError("error")
            self.worker.clean_workdir()

    async def assert_logging_setup(
        self,
        argv_setup: dict[str, Any],
        config_setup: dict[str, Any],
        expected: dict[str, Any],
    ) -> None:
        """
        Assert logging setup for argv_setup and config_setup.

        :param argv_setup: parameters that a user might have passed using
          the CLI
        :param config_setup: parameters that might be set in the config.ini
        :param expected: given the argv_setup and config_setup: what is the
          expected configuration
        """
        if config_setup['log-level'] is not None:
            self.config['General']['log-level'] = logging.getLevelName(
                config_setup['log-level']
            )

        # Normalizes log_level: always using the str representation and not
        # the int
        if argv_setup['log_level'] is not None:
            argv_setup['log_level'] = logging.getLevelName(
                argv_setup['log_level']
            )
        expected['level'] = logging.getLevelName(expected['level'])

        # Patches basicConfig
        patcher = mock.patch(
            'debusine.worker._worker.logging.basicConfig', autospec=True
        )
        mocked_logging = patcher.start()
        self.addCleanup(patcher.stop)

        await self.setup_worker(**argv_setup)

        self.assertTrue(mocked_logging.called)
        called_args = mocked_logging.call_args[1]
        self.assertEqual(called_args['level'], expected['level'])

        # Verify that handlers include JournalHandler
        self.assertIn('handlers', called_args)
        handlers = called_args['handlers']
        self.assertEqual(len(handlers), 1)
        handler_class = type(handlers[0]).__name__
        self.assertIn("JournalHandler", handler_class)

    async def test_setup_logging_no_logging_configuration(self) -> None:
        """Logging system: no explicit setup: use Worker.DEFAULT_LOG_LEVEL."""
        argv_setup = {'log_level': None}
        config_setup = {'log-level': None}
        expected = {'level': Worker.DEFAULT_LOG_LEVEL}
        await self.assert_logging_setup(argv_setup, config_setup, expected)

    async def test_setup_logging_from_constructor(self) -> None:
        """Logging system: use settings from the Worker constructor."""
        argv_setup = {'log_level': logging.DEBUG}
        config_setup = {'log-level': None}
        expected = {'level': logging.DEBUG}

        await self.assert_logging_setup(argv_setup, config_setup, expected)

    async def test_setup_logging_from_config(self) -> None:
        """Logging system: use settings from the configuration file."""
        argv_setup = {'log_level': None}
        config_setup = {'log-level': logging.DEBUG}
        expected = {'level': logging.DEBUG}
        await self.assert_logging_setup(argv_setup, config_setup, expected)

    async def test_setup_logging_worker_constructor_overwrites_config(
        self,
    ) -> None:
        """Logging system: use settings from the Worker constructor."""
        argv_setup = {'log_level': logging.DEBUG}
        config_setup = {'log-level': logging.ERROR}
        expected = {'level': logging.DEBUG}

        await self.assert_logging_setup(argv_setup, config_setup, expected)

    async def test_log_forced_exit(self) -> None:
        """log_forced_exit logs the event."""
        await self.setup_worker()
        assert self.worker is not None

        with self.assertLogsContains(
            "Terminated with signal SIGINT",
            logger="debusine.worker",
            level=logging.INFO,
        ):
            self.worker.log_forced_exit(signal.SIGINT)

    async def test_asyncio_sleep(self) -> None:
        """Worker._asyncio_sleep calls asyncio.sleep."""
        mocked = self.enterContext(mock.patch('asyncio.sleep', autospec=True))
        await Worker._asyncio_sleep(1.5)
        mocked.assert_called_with(1.5)

    async def test_create_task_signal_int_term_handler(self) -> None:
        """
        Test Task returned from Worker._create_task_signal_int_term_handler.

        Assert that the task logs "Terminated with signal ...".
        Other tests on the behaviour of Worker._signal_int_term_handler
        is implemented in the method test_signal_int_term_handler.

        Testing that Worker._create_task_signal_int_term_handler returns
        a Task that will execute Worker._signal_int_term_handler might be
        possible but only if accessing Task._coro private method.
        """
        await self.setup_worker()
        assert self.worker is not None

        test_signals = [signal.SIGINT, signal.SIGTERM]

        for test_signal in test_signals:
            with self.subTest(test_signal=repr(test_signal)):
                sig_handler = self.worker._create_task_signal_int_term_handler(
                    test_signal
                )

                self.assertIsInstance(sig_handler, asyncio.Task)

                with (
                    mock.patch.object(
                        self.worker.event_queue, "put"
                    ) as queue_put,
                    self.assertLogsContains(
                        f"Terminated with signal {test_signal.name}",
                        logger="debusine.worker",
                    ),
                ):
                    await sig_handler
                    queue_put.assert_called_once()
                    self.assertEqual(
                        str(queue_put.call_args.args[0]),
                        f"shutdown: signal {test_signal} received",
                    )

    async def test_setup_signal_int_term_handler(self) -> None:
        """Test that SIGINT/SIGTERM setup is done in Worker.main()."""
        await self.setup_worker()
        assert self.worker is not None

        connect = self.enterContext(
            mock.patch.object(
                self.worker,
                "process_notifications",
                return_value=None,
            )
        )

        add_signal_handler = self.enterContext(
            mock.patch.object(asyncio.get_running_loop(), "add_signal_handler")
        )
        with self.assertRaisesSystemExit(4):
            await self.worker.main()

        add_signal_handler.assert_has_calls(
            [
                mock.call(signal.SIGINT, mock.ANY),
                mock.call(signal.SIGTERM, mock.ANY),
            ]
        )

        connect.assert_awaited()

    async def test_int_term_handler_unexpected_signum(self) -> None:
        await self.setup_worker()
        assert self.worker is not None
        await self.worker._signal_int_term_handler(signal.SIGUSR1)
        self.assertQueuedEvents(["shutdown: signal 10 received"])

    async def test_exit_code(self) -> None:
        await self.setup_worker()
        assert self.worker is not None
        self.enterContext(mock.patch.object(self.worker, "require_token"))
        process_notifications = self.enterContext(
            mock.patch.object(
                self.worker,
                "process_notifications",
            )
        )
        for event, code in (
            (_worker.ServerDisconnectShutdownEvent(), 1),
            (_worker.ReportTaskResultFailedShutdownEvent(), 2),
            (
                _worker.SignalReceivedShutdownEvent(signal.SIGTERM),
                128 + signal.SIGTERM,
            ),
            (
                _worker.SignalReceivedShutdownEvent(signal.SIGINT),
                128 + signal.SIGINT,
            ),
            (
                _worker.SignalReceivedShutdownEvent(signal.SIGUSR1),
                128 + signal.SIGUSR1,
            ),
            (None, 4),
        ):
            with self.subTest(event=str(event)):
                process_notifications.return_value = event
                with self.assertRaisesSystemExit(code):
                    await self.worker.main()


class WorkerEventsTests(WorkerTestCase):
    """Test worker event handling."""

    worker: Worker

    @override
    async def setUpAsync(self) -> None:
        await super().setUpAsync()

        token = '6c931875627131b5135b7de3371c44'
        self.server_config.registered_token = token
        self.config.write_token(token)

        # Server notification payloads that will be generated by the mocked
        # ServerNotifications
        self.server_notification_payloads: list[dict[str, Any]] = []

        # Mock the server notification reader
        async def mock_messages() -> AsyncGenerator[dict[str, Any]]:
            for payload in self.server_notification_payloads:
                yield payload

        self.server_notifications = mock.MagicMock(spec=ServerNotifications)
        self.server_notifications.__aenter__.return_value = (
            self.server_notifications
        )
        self.server_notifications.__aexit__.return_value = None
        self.server_notifications.messages = mock_messages

        self.enterContext(
            mock.patch(
                "debusine.client.debusine.Debusine.server_notifications",
                return_value=self.server_notifications,
            )
        )

        self.worker = await self.enterAsyncContext(
            Worker(
                worker_type=WorkerType.EXTERNAL,
                working_directory=self.temp_config_directory.as_posix(),
                config=self.config,
            )
        )

        # Don't risk triggering local hooks
        self.worker.AFTER_WORK_REQUEST_HOOK_DIR = Path("/nonexistent")

    def assertQueuedEvents(self, expected: list[str]) -> None:
        """
        Check that the queued events match the given list.

        Events are stringified before comparison
        """
        actual: list[str] = []
        while not self.worker.event_queue.empty():
            evt = self.worker.event_queue.get_nowait()
            assert isinstance(evt, _worker.Event)
            actual.append(str(evt))
        expected = [
            *expected,
            "shutdown: server notification connection closed",
        ]
        assert actual == expected

    async def test_handle_server_notifications_payloads_to_events(self) -> None:
        RESPONSES = ServerConfig.RESPONSES
        self.server_notification_payloads = [
            RESPONSES["connected"],
            RESPONSES["request_dynamic_metadata"],
            RESPONSES["work_request_available"],
        ]
        await self.worker.handle_server_notifications()
        self.assertQueuedEvents(
            [
                "connected",
                "request_dynamic_metadata",
                "work_request_available",
            ]
        )

    async def test_handle_server_notifications_ignore_unsupported_payload(
        self,
    ) -> None:
        self.server_notification_payloads = [
            {"text": "unsupported"},
            ServerConfig.RESPONSES["connected"],
        ]
        with self.assertLogs(
            logger="debusine.worker", level=logging.WARNING
        ) as log:
            await self.worker.handle_server_notifications()
        self.assertQueuedEvents(["connected"])
        self.assertEqual(
            log.output,
            [
                "WARNING:debusine.worker:"
                "Ignoring unsupported or malformed server message:"
                " {'text': 'unsupported'}: unsupported server event"
            ],
        )

    async def test_handle_server_notifications_ignore_payload_without_text(
        self,
    ) -> None:
        self.server_notification_payloads = [
            {},
            ServerConfig.RESPONSES["connected"],
        ]
        with self.assertLogs(
            logger="debusine.worker", level=logging.WARNING
        ) as log:
            await self.worker.handle_server_notifications()
        self.assertQueuedEvents(["connected"])
        self.assertEqual(
            log.output,
            [
                "WARNING:debusine.worker:"
                "Ignoring unsupported or malformed server message:"
                " {}: message payload without 'text'"
            ],
        )

    async def test_process_notifications_shutdown(self) -> None:
        event = _worker.SignalReceivedShutdownEvent(signal.SIGUSR1)
        await self.worker.event_queue.put(event)
        with self.assertLogs(
            logger="debusine.worker", level=logging.INFO
        ) as log:
            self.assertEqual(await self.worker.process_notifications(), event)
        self.assertEqual(
            log.output,
            [
                "INFO:debusine.worker:"
                "Shutdown requested: shutdown: signal 10 received"
            ],
        )

    async def test_process_notifications_connected(self) -> None:
        await self.worker.event_queue.put(_worker.ConnectedEvent())
        with self.assertLogs(
            logger="debusine.worker", level=logging.DEBUG
        ) as log:
            await self.worker.process_notifications()
        self.assertEqual(
            log.output[:2],
            [
                "DEBUG:debusine.worker:Worker event: connected",
                "DEBUG:debusine.worker:Received the 'connected' message",
            ],
        )

    async def test_process_notifications_request_dynamic_metadata(self) -> None:
        await self.worker.event_queue.put(_worker.RequestDynamicMetadataEvent())
        sdm = self.enterContext(
            mock.patch.object(self.worker, "_send_dynamic_metadata")
        )
        await self.worker.process_notifications()
        sdm.assert_called()

    async def test_process_notifications_request_dynamic_metadata_awaited(
        self,
    ) -> None:
        await self.worker.event_queue.put(_worker.RequestDynamicMetadataEvent())

        exception = RuntimeError("expected error")

        async def mock_send_dynamic_metadata() -> None:
            raise exception

        self.enterContext(
            mock.patch.object(
                self.worker,
                "_send_dynamic_metadata",
                side_effect=mock_send_dynamic_metadata,
            )
        )
        with self.assertRaises(ExceptionGroup) as exc:
            await self.worker.process_notifications()
        self.assertIs(exc.exception.exceptions[0], exception)

    async def test_process_notifications_work_request_available(self) -> None:
        await self.worker.event_queue.put(_worker.WorkRequestAvailableEvent())
        sdm = self.enterContext(
            mock.patch.object(self.worker, "_request_and_execute_work_request")
        )
        await self.worker.process_notifications()
        sdm.assert_called()

    async def test_process_notifications_work_request_available_awaited(
        self,
    ) -> None:
        await self.worker.event_queue.put(_worker.WorkRequestAvailableEvent())

        exception = RuntimeError("expected error")

        async def mock_request_and_execute_work_request() -> None:
            raise exception

        self.enterContext(
            mock.patch.object(
                self.worker,
                "_request_and_execute_work_request",
                side_effect=mock_request_and_execute_work_request,
            )
        )
        with self.assertRaises(ExceptionGroup) as exc:
            await self.worker.process_notifications()
        self.assertIs(exc.exception.exceptions[0], exception)

    async def test_process_notifications_unsupported(self) -> None:
        await self.worker.event_queue.put("invalid")  # type: ignore[arg-type]
        with (
            self.assertLogs(
                logger="debusine.worker", level=logging.WARNING
            ) as log,
        ):
            await self.worker.process_notifications()
        self.assertEqual(
            log.output,
            ["WARNING:debusine.worker:Ignoring unsupported event: invalid"],
        )


class WorkerTests(WorkerTestCase):
    """Test Worker class."""

    worker: Worker

    @override
    async def setUpAsync(self) -> None:
        await super().setUpAsync()

        token = '6c931875627131b5135b7de3371c44'
        self.server_config.registered_token = token
        self.config.write_token(token)

        self.worker = await self.enterAsyncContext(
            Worker(
                working_directory=self.temp_config_directory.as_posix(),
                worker_type=WorkerType.EXTERNAL,
                config=self.config,
            )
        )

        # Don't risk triggering local hooks
        self.worker.AFTER_WORK_REQUEST_HOOK_DIR = Path("/nonexistent")

    def set_events(
        self, *events: _worker.Event, include_quit_main_loop: bool = True
    ) -> None:
        self.enterContext(
            mock.patch.object(self.worker, "handle_server_notifications")
        )
        for event in events:
            self.worker.event_queue.put_nowait(event)
        if include_quit_main_loop:
            self.worker.event_queue.put_nowait(_worker.QuitMainLoopEvent())

    def patch_worker(self, attribute: str, **kwargs: Any) -> mock.AsyncMock:
        """Patch an attribute of the worker object."""
        patcher = mock.patch.object(self.worker, attribute, autospec=True)
        mocked = patcher.start()
        for key, value in kwargs.items():
            setattr(mocked, key, value)
        self.addCleanup(patcher.stop)
        return mocked

    def patch_debusine_work_request_completed_update(self) -> MagicMock:
        """Patch Debusine.work_request_completed_update. Return its mock."""
        # setup_worker must be called first.
        assert self.worker is not None
        return self.enterContext(
            mock.patch(
                "debusine.client.debusine.Debusine.work_request_completed_update",
                autospec=True,
            )
        )

    def assertQueuedEvents(self, expected: list[str]) -> None:
        """
        Check that the queued events match the given list.

        Events are stringified before comparison
        """
        actual: list[str] = []
        while not self.worker.event_queue.empty():
            evt = self.worker.event_queue.get_nowait()
            assert isinstance(evt, _worker.Event)
            actual.append(str(evt))
        assert actual == expected

    async def capture_build_task(self) -> asyncio.Task[None]:
        """Wait for a build task to start, and return it."""
        orig_task_group_create_task = asyncio.TaskGroup.create_task

        # Monitor tasks created during event processing, to capture the first
        # that runs Worker._request_and_execute_work_request
        build_task: asyncio.Task[None] | None = None

        found = asyncio.Event()

        def spy_create_task(
            task_group: asyncio.TaskGroup, awaitable: Coroutine[Any, Any, Never]
        ) -> asyncio.Task[Any]:
            nonlocal build_task
            task = orig_task_group_create_task(task_group, awaitable)
            if (
                awaitable.__qualname__
                == "Worker._request_and_execute_work_request"
            ):
                build_task = task
                found.set()
            return task

        with mock.patch(
            "asyncio.TaskGroup.create_task",
            side_effect=spy_create_task,
            autospec=True,
        ):
            await found.wait()

        assert build_task is not None
        return build_task

    async def test_process_message_dynamic_metadata_request_signing(
        self,
    ) -> None:
        """Signing workers send the correct worker type in dynamic metadata."""
        self.worker._worker_type = WorkerType.SIGNING

        with (
            mock.patch(
                "debusine.worker._worker.analyze_worker_all_executors",
                return_value={},
            ),
            mock.patch(
                "debusine.worker._worker.analyze_external_worker_tasks",
                return_value={},
            ),
            mock.patch(
                "debusine.worker._worker.Worker._get_provided_tags",
                autospec=True,
                return_value=set(),
            ),
        ):
            await self.worker._send_dynamic_metadata()

        assert self.server_latest_request is not None
        self.assertEqual(
            self.server_latest_request.path, "/api/1.0/worker/dynamic-metadata/"
        )
        self.assertTrue(self.server_latest_request.is_authenticated)
        self.assertEqual(
            self.server_latest_request.json_content,
            {
                **system_metadata(task_models.WorkerType.SIGNING),
                "provided_tags": [],
            },
        )

    async def test_send_dynamic_metadata(self) -> None:
        """Test a successful dynamic metadata update."""
        worker_metadata = {"test": True}

        with (
            mock.patch(
                "debusine.worker._worker.Worker._get_worker_metadata",
                autospec=True,
                return_value=worker_metadata,
            ),
            self.assertLogs(
                logger="debusine.worker", level=logging.DEBUG
            ) as logs,
        ):
            await self.worker._send_dynamic_metadata()

        regex = r'Send dynamic metadata'
        found = sum(
            int(re.search(regex, log) is not None) for log in logs.output
        )
        self.assertEqual(found, 1)

        assert self.server_latest_request is not None
        self.assertEqual(
            self.server_latest_request.path, '/api/1.0/worker/dynamic-metadata/'
        )

        self.assertTrue(self.server_latest_request.is_authenticated)

        self.assertEqual(
            self.server_latest_request.json_content, worker_metadata
        )

    async def test_send_dynamic_metadata_failed(self) -> None:
        """Test a failed dynamic metadata update."""
        executors_metadata = {"executor:unshare:available": True}
        tasks_metadata = {"sbuild:version": 1, "sbuild:available": True}

        self.server_config.dynamic_metadata_status_code = (
            HTTPServiceUnavailable.status_code
        )

        with (
            mock.patch(
                "debusine.worker._worker.analyze_worker_all_executors",
                autospec=True,
                return_value=executors_metadata,
            ),
            mock.patch(
                'debusine.worker._worker.analyze_external_worker_tasks',
                autospec=True,
                return_value=tasks_metadata,
            ),
            self.assertLogsContains(
                "ERROR:debusine.worker:"
                "Failed to send dynamic_metadata: (response: 503): ",
                logger="debusine.worker",
            ),
        ):
            await self.worker._send_dynamic_metadata()

        assert self.server_latest_request is not None
        self.assertEqual(
            self.server_latest_request.path, '/api/1.0/worker/dynamic-metadata/'
        )

    async def receive_task_submit_result(
        self,
        execute_result: WorkRequestResults,
        json_result: str,
    ) -> None:
        """Worker fetches the task and validates submitted JSON."""
        runner_execute = self.enterContext(
            mock.patch(
                "debusine.worker.task_runner.TaskRunner.execute",
                return_value=execute_result,
            )
        )

        # No request to any URL...
        self.assertIsNone(self.server_latest_request)

        self.set_events(_worker.WorkRequestAvailableEvent())

        work_request_completed_mocked = (
            self.patch_debusine_work_request_completed_update()
        )

        with self.assertRaisesSystemExit(4):
            await self.worker.main()

        # The WorkRequest is executed...
        self.assertEqual(runner_execute.call_count, 1)

        work_request_completed_mocked.assert_called_once_with(
            mock.ANY, self.work_request.id, json_result, mock.ANY
        )

        output_data = work_request_completed_mocked.call_args.args[3]
        self.assertIsInstance(output_data, OutputData)
        self.assertIsNotNone(output_data.runtime_statistics)

    async def test_task_is_received_executed_submit_success(self) -> None:
        """Receive a task, execute it and submit success completion."""
        await self.receive_task_submit_result(
            WorkRequestResults.SUCCESS, "success"
        )

    async def test_task_is_received_executed_submit_failure(self) -> None:
        """Receive a task, executes it (failure) and submits failure."""
        await self.receive_task_submit_result(
            WorkRequestResults.FAILURE, "failure"
        )

    async def test_print_no_work_request_available(self) -> None:
        """Worker logs 'No work request available'."""
        self.set_events(_worker.WorkRequestAvailableEvent())
        self.server_config.get_next_for_worker = (
            ServerConfig.GetNextForWorkerResponse.NO_WORK_REQUEST_PENDING
        )

        log_message = 'No work request available'

        with self.assertLogsContains(
            log_message, logger="debusine.worker", level=logging.DEBUG
        ):
            await self.worker.process_notifications()

    async def test_log_work_request_ignored_already_running(self) -> None:
        """Worker receive two Work Request available, fetches only one."""
        wait_until_requesting = asyncio.Event()
        wait_until_condition_tested = asyncio.Event()

        async def _request_work_request() -> None:
            wait_until_requesting.set()
            await wait_until_condition_tested.wait()

        self.enterContext(
            mock.patch.object(
                self.worker,
                "_request_work_request",
                side_effect=_request_work_request,
            )
        )

        with self.assertLogsContains(
            "Waited 0 seconds for the end of the former task",
            logger="debusine.worker",
            level=logging.DEBUG,
        ):
            first_request = asyncio.create_task(
                self.worker._request_and_execute_work_request()
            )
            await wait_until_requesting.wait()
        # The first work request now has the lock

        # Log that a Work Request is available (coming from the server)
        # but the Worker ignores it because the worker is already
        # executing a task
        with self.assertLogsContains(
            "Worker is busy and can't execute a new work request",
            logger="debusine.worker",
        ):
            self.patch_worker("_asyncio_sleep")
            await self.worker._request_and_execute_work_request()

        wait_until_condition_tested.set()
        with self.assertLogsContains(
            "No work request available",
            logger="debusine.worker",
            level=logging.INFO,
        ):
            await first_request

    async def test_shutdown_event_aborts_task(self) -> None:
        # Reduce wait_popen polling interval to make the test faster
        orig_wait_popen = Sbuild._wait_popen

        def fast_wait_popen(
            self: Sbuild,
            popen: subprocess.Popen[AnyStr],
            timeout: float,  # noqa: ARG001
        ) -> int:
            return orig_wait_popen(self, popen, 0.05)

        self.enterContext(
            mock.patch(
                "debusine.tasks.Sbuild._wait_popen",
                side_effect=fast_wait_popen,
                autospec=True,
            )
        )

        loop = asyncio.get_running_loop()
        execute_started = asyncio.Event()
        execute_done = asyncio.Event()

        runner_shutdown = self.enterContext(
            self.spy("debusine.worker.task_runner.TaskRunner.shutdown")
        )

        def mock_execute(self: Sbuild) -> WorkRequestResults:
            nonlocal execute_started
            loop.call_soon_threadsafe(execute_started.set)
            self.run_cmd(["sleep", "inf"], working_directory=Path.cwd())
            loop.call_soon_threadsafe(execute_done.set)
            return WorkRequestResults.SUCCESS

        self.enterContext(
            mock.patch(
                "debusine.tasks.Sbuild.execute_logging_exceptions",
                side_effect=mock_execute,
                autospec=True,
            )
        )

        report_task_result = self.enterContext(
            mock.patch.object(self.worker, "report_task_result")
        )

        self.set_events(
            _worker.WorkRequestAvailableEvent(), include_quit_main_loop=False
        )

        async with asyncio.TaskGroup() as tg:
            capture_build_task = tg.create_task(self.capture_build_task())
            tg.create_task(self.worker.process_notifications())
            await execute_started.wait()
            await self.worker.event_queue.put(
                _worker.SignalReceivedShutdownEvent(signal.SIGTERM)
            )

        build_task = capture_build_task.result()
        self.assertTrue(build_task.cancelled())
        report_task_result.assert_not_called()
        runner_shutdown.assert_called()

        runner: TaskRunner = runner_shutdown.call_args.args[0]
        assert runner.task is not None
        self.assertTrue(runner.task.aborted)

        await execute_done.wait()

    async def test_signal_int_term_handler_schedules_shutdown(self) -> None:
        self.enterContext(mock.patch.object(self.worker, "require_token"))

        self.assertTrue(self.worker.event_queue.empty())
        with self.assertLogsContains(
            "Terminated with signal SIGINT", logger="debusine.worker"
        ):
            await self.worker._signal_int_term_handler(signal.SIGINT)
        self.assertQueuedEvents(["shutdown: signal 2 received"])

    async def assert_invalid_get_next_for_worker_logging(
        self, get_next_for_worker: ServerConfig.GetNextForWorkerResponse
    ) -> None:
        """Assert log_message is logged for get_next_for_worker response."""
        self.set_events(_worker.WorkRequestAvailableEvent())
        self.server_config.get_next_for_worker = get_next_for_worker

        assert self.worker is not None

        with self.assertLogsContains(
            "Invalid WorkRequest received from",
            logger="debusine.worker",
            level=logging.WARNING,
        ):
            await self.worker.process_notifications()

    async def test_request_work_request_invalid_work_request_response(
        self,
    ) -> None:
        """Worker logs 'Invalid content of...'."""
        await self.assert_invalid_get_next_for_worker_logging(
            ServerConfig.GetNextForWorkerResponse.INVALID_WORK_REQUEST
        )

    async def test_request_work_request_invalid_json_response(self) -> None:
        """Worker logs 'Invalid JSON response...'."""
        await self.assert_invalid_get_next_for_worker_logging(
            ServerConfig.GetNextForWorkerResponse.INVALID_JSON
        )

    async def test_request_work_request_retries(self) -> None:
        """_request_work_request retires failed requests."""
        self.server_config.get_next_for_worker = Response(
            status=HTTPServiceUnavailable.status_code
        )

        def sleep(delay: float) -> None:  # noqa: ARG001, U100
            self.server_config.get_next_for_worker = (
                ServerConfig.GetNextForWorkerResponse.VALID_WORK_REQUEST
            )

        self.patch_worker("_asyncio_sleep", side_effect=sleep)
        response = await self.worker._request_work_request()
        self.assertIsNotNone(response)

    async def assert_execute_work_request_send_and_log_error(
        self,
        error_msg: str,
        error_code: str,
        worker_type: Literal[
            WorkerType.EXTERNAL, WorkerType.SIGNING
        ] = WorkerType.EXTERNAL,
    ) -> None:
        """Assert error is sent to the server and logged."""
        self.worker._worker_type = worker_type

        work_request_completed_mocked = (
            self.patch_debusine_work_request_completed_update()
        )

        with self.assertLogsContains(error_msg, logger="debusine.worker"):
            await self.worker._request_and_execute_work_request()

        work_request_completed_mocked.assert_called_once_with(
            mock.ANY, mock.ANY, "error", mock.ANY
        )
        _, work_request_id, _, output_data = (
            work_request_completed_mocked.call_args[0]
        )
        self.assertIsInstance(work_request_id, int)
        self.assertIsInstance(output_data, OutputData)
        self.assertEqual(
            output_data.errors,
            [OutputDataError(message=error_msg, code=error_code)],
        )

    async def test_task_configure_error_submit_error(self) -> None:
        """Sbuild.configure fails, Worker submits error."""
        self.enterContext(
            mock.patch.object(
                Sbuild,
                '_configure',
                autospec=True,
                side_effect=TaskConfigError('Invalid schema'),
            )
        )
        await self.assert_execute_work_request_send_and_log_error(
            "Task: sbuild Error configure: Invalid schema", "configure-failed"
        )

    async def test_task_execution_error_submit_error(self) -> None:
        """Sbuild.execute raises TaskConfigError, Worker submits error."""
        self.enterContext(
            mock.patch.object(
                Sbuild,
                'execute',
                autospec=True,
                side_effect=TaskConfigError('Broken'),
            )
        )

        await self.assert_execute_work_request_send_and_log_error(
            'Task: sbuild Error execute: Broken', "execute-failed"
        )

    def assert_functools_partial_equal(
        self, partial1: functools.partial[Any], partial2: functools.partial[Any]
    ) -> None:
        """Assert two functools.partials are semantically equal."""
        # functools.partial does not implement __eq__
        self.assertEqual(partial1.func, partial2.func)
        self.assertEqual(partial1.args, partial2.args)
        self.assertEqual(partial1.keywords, partial2.keywords)

    async def test_main(self) -> None:
        require_token = self.patch_worker("require_token")
        process_notifications = self.patch_worker(
            "process_notifications", return_value=None
        )

        mocked_add_signal_handler = self.enterContext(
            mock.patch.object(
                asyncio.get_running_loop(),
                "add_signal_handler",
                autospec=True,
            )
        )

        with (
            self.assertLogsContains(
                "Debusine worker is starting. Version:",
                logger="debusine.worker",
                level=logging.INFO,
            ),
            self.assertRaisesSystemExit(4),
        ):
            await self.worker.main()

        sigint_call = mocked_add_signal_handler.mock_calls[0][1]
        self.assertEqual(sigint_call[0], signal.SIGINT)
        self.assert_functools_partial_equal(
            sigint_call[1],
            functools.partial(
                self.worker._create_task_signal_int_term_handler, signal.SIGINT
            ),
        )

        sigterm_call = mocked_add_signal_handler.mock_calls[1][1]
        self.assertEqual(sigterm_call[0], signal.SIGTERM)
        self.assert_functools_partial_equal(
            sigterm_call[1],
            functools.partial(
                self.worker._create_task_signal_int_term_handler, signal.SIGTERM
            ),
        )

        self.assertEqual(len(mocked_add_signal_handler.mock_calls), 2)

        require_token.assert_awaited()
        process_notifications.assert_awaited()

    async def test_report_task_result_network_error(self) -> None:
        work_request_completed_mocked = (
            self.patch_debusine_work_request_completed_update()
        )
        work_request_completed_mocked.side_effect = Exception("network error")

        async with TaskRunner(
            self.worker._config,
            self.worker._worker_type,
            self.work_request,
            self.worker._workdir,
        ) as runner:
            with (
                self.assertLogsContains(
                    "Cannot reach server",
                    logger="debusine.worker",
                    level=logging.ERROR,
                ),
            ):
                await self.worker.report_task_result(
                    runner,
                    WorkRequestResults.ERROR,
                    OutputData(
                        errors=[
                            OutputDataError(
                                message="Expected error", code="expected-error"
                            )
                        ]
                    ),
                )

        self.assertQueuedEvents(["shutdown: cannot send results to server"])

    async def test_report_task_result_do_not_send_aborted(self) -> None:
        work_request_completed_update = (
            self.patch_debusine_work_request_completed_update()
        )

        async with TaskRunner(
            self.worker._config,
            self.worker._worker_type,
            self.work_request,
            self.worker._workdir,
        ) as runner:
            await runner.setup_task()
            assert runner.task is not None
            runner.task.abort()
            with self.assertLogs(
                logger="debusine.worker", level=logging.INFO
            ) as log:
                await self.worker.report_task_result(
                    runner, WorkRequestResults.ERROR, OutputData()
                )

        self.assertEqual(
            log.output, ["INFO:debusine.worker:Task: sbuild has been aborted"]
        )
        work_request_completed_update.assert_not_called()

    async def test_report_task_result_runs_after_work_request_hook(
        self,
    ) -> None:
        async with TaskRunner(
            self.worker._config,
            self.worker._worker_type,
            self.work_request,
            self.worker._workdir,
        ) as runner:
            await runner.setup_task()
            assert runner.task is not None
            runner.task.abort()

            with self.assertLogs(
                logger="debusine.worker", level=logging.DEBUG
            ) as log:
                await self.worker.report_task_result(
                    runner, WorkRequestResults.ERROR, OutputData()
                )

        self.assertEqual(
            log.output,
            [
                "INFO:debusine.worker:Task: sbuild has been aborted",
                "DEBUG:debusine.worker:"
                "No hooks in '/nonexistent', skipping execution",
            ],
        )

    async def test_run_after_work_request_hook(self) -> None:
        """Run a simple script via _run_after_work_request_hook."""
        hook_dir = Path(self.create_temporary_directory())
        self.worker.AFTER_WORK_REQUEST_HOOK_DIR = hook_dir

        marker = hook_dir / "marker"
        script = hook_dir / "hook"
        script.write_text(
            dedent(
                f"""\
                #!/bin/sh
                touch {marker}
                """
            )
        )
        script.chmod(0o755)
        await self.worker._run_after_work_request_hook()
        self.assertTrue(marker.exists())

    async def test_run_after_work_request_hook_error(self) -> None:
        """Run a simple failing script via _run_after_work_request_hook."""
        hook_dir = Path(self.create_temporary_directory())
        self.worker.AFTER_WORK_REQUEST_HOOK_DIR = hook_dir

        script = hook_dir / "hook"
        script.write_text(
            dedent(
                """\
                #!/bin/sh
                exit 1
                """
            )
        )
        script.chmod(0o755)
        with self.assertLogsContains(
            f"Hooks failed to execute in {str(hook_dir)!r} .\n"
            "stdout:\n"
            "\n"
            "stderr:\n"
            f"run-parts: {hook_dir}/hook exited with return code 1",
            logger="debusine.worker",
            level=logging.WARNING,
        ):
            await self.worker._run_after_work_request_hook()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for OpenMetrics statistics in Debusine."""

import datetime as dt

from django.test import SimpleTestCase, override_settings
from prometheus_client.core import Metric

from debusine.artifacts import LocalArtifact
from debusine.artifacts.models import ArtifactCategory, TaskTypes
from debusine.assets.models import AssetCategory
from debusine.db.models import (
    Artifact,
    Identity,
    User,
    WorkRequest,
    WorkerPoolStatistics,
)
from debusine.server.open_metrics import DebusineCollector, ZeroGauge
from debusine.server.signon.providers import Provider
from debusine.tasks.models import BackendType, NoopData, WorkerType
from debusine.test.django import TestCase


class ZeroGaugeTests(SimpleTestCase):
    """Tests for our zero-filling ZeroGauge."""

    def test_add_metric_stores_samples(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=["type"])
        gauge.add_metric(["foo"], 1)

        self.assertEqual(len(gauge.samples), 1)
        sample = gauge.samples[0]
        self.assertEqual(sample.value, 1)
        self.assertEqual(sample.labels, {"type": "foo"})

    def test_add_metric_sets_seen(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=["type"])
        gauge.add_metric(["foo"], 1)
        gauge.add_metric(["bar"], 2)

        self.assertEqual(gauge._seen, {("foo",), ("bar",)})

    def test_fill_zeros(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=["type"])
        gauge.add_metric(["foo"], 1)
        gauge.fill_zeros(("foo", "bar"))

        self.assertEqual(len(gauge.samples), 2)
        sample = gauge.samples[1]
        self.assertEqual(sample.value, 0)
        self.assertEqual(sample.labels, {"type": "bar"})

    def test_fill_zeros_complete(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=["type"])
        gauge.add_metric(["foo"], 1)
        gauge.add_metric(["bar"], 2)
        gauge.fill_zeros(("foo", "bar"))

        self.assertEqual(len(gauge.samples), 2)
        self.assertNotEqual(gauge.samples[0].value, 0)
        self.assertNotEqual(gauge.samples[1].value, 0)

    def test_fill_zeros_empty(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=["type"])
        gauge.fill_zeros(("foo", "bar"))

        self.assertEqual(len(gauge.samples), 2)
        self.assertEqual(gauge.samples[0].value, 0)
        self.assertEqual(gauge.samples[1].value, 0)

    def test_fill_zeros_matrix(self) -> None:
        gauge = ZeroGauge("test", "Description", labels=("a", "b"))
        gauge.add_metric(("0", "0"), 1)
        gauge.fill_zeros(("0", "1"), ("0", "1"))

        self.assertEqual(len(gauge.samples), 4)


class DebusineCollectorTests(TestCase):
    """Test DebusineCollector's collectors."""

    def collector(self) -> DebusineCollector:
        collector = DebusineCollector()
        return collector

    def sample_by_label(self, metric: Metric, labels: dict[str, str]) -> float:
        """Find the sample for labels in metric."""
        for sample in metric.samples:
            if sample.labels == labels:
                return sample.value
        raise KeyError(f"{labels} not found in {metric}")

    def assertMetricEquals(
        self, metric: Metric, value: float, **labels: str
    ) -> None:
        """Assert that there is a sample for labels in metric with value."""
        self.assertEqual(self.sample_by_label(metric, labels), value)

    def assertNoMetric(self, metric: Metric, **labels: str) -> None:
        """Assert that there is no sample for labels in metric."""
        with self.assertRaises(KeyError):
            self.sample_by_label(metric, labels)

    def assertNoNullLabels(self, metric: Metric) -> None:
        """Assert that all metric labels have string values."""
        for sample in metric.samples:
            for key in sample.labels:
                self.assertIsNotNone(
                    sample.labels[key], msg=f"Null value for label '{key}'"
                )

    def test_count_by_category(self) -> None:
        self.playground.create_artifact()

        gauge = self.collector().count_by_category(
            Artifact, LocalArtifact.artifact_categories()
        )
        self.assertEqual(gauge.name, "artifacts")
        self.assertEqual(gauge.documentation, "Number of Artifacts")

        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(
            gauge, 1, category=ArtifactCategory.TEST, scope=scope
        )
        self.assertMetricEquals(
            gauge, 0, category=ArtifactCategory.BLHC, scope=scope
        )

    def test_count_assets(self) -> None:
        self.playground.create_signing_key_asset()

        gauge = self.collector().count_assets()
        self.assertEqual(gauge.name, "assets")

        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(
            gauge, 1, category=AssetCategory.SIGNING_KEY, scope=scope
        )
        self.assertNoMetric(
            gauge, category=AssetCategory.CLOUD_PROVIDER_ACCOUNT, scope=scope
        )
        self.assertMetricEquals(
            gauge, 0, category=AssetCategory.CLOUD_PROVIDER_ACCOUNT, scope=""
        )
        self.assertNoMetric(gauge, category=AssetCategory.SIGNING_KEY, scope="")

    def test_measure_file_stores(self) -> None:
        store = self.playground.get_default_file_store()
        contents = b"abc123"
        self.playground.create_file_in_backend(contents=contents)
        gauges = {
            gauge.name: gauge
            for gauge in self.collector().measure_file_stores()
        }
        self.assertEqual(len(gauges), 2)

        file_store_size = gauges["file_store_size"]
        self.assertMetricEquals(
            file_store_size,
            len(contents),
            backend=store.backend,
            name=store.name,
        )
        file_store_max_size = gauges["file_store_max_size"]
        self.assertNoMetric(
            file_store_max_size, backend=store.backend, name=store.name
        )

    def test_measure_file_stores_max_size(self) -> None:
        store = self.playground.get_default_file_store()
        store.max_size = 1024
        store.save()
        gauges = {
            gauge.name: gauge
            for gauge in self.collector().measure_file_stores()
        }
        file_store_max_size = gauges["file_store_max_size"]
        self.assertMetricEquals(
            file_store_max_size, 1024, backend=store.backend, name=store.name
        )

    def test_count_groups(self) -> None:
        scope1 = self.playground.get_default_scope()
        scope2 = self.playground.get_or_create_scope("scope2")
        self.playground.create_group("group1", scope=scope1)
        self.playground.create_group("group2", scope=scope2, ephemeral=True)
        gauge = self.collector().count_groups()
        self.assertEqual(gauge.name, "groups")
        self.assertMetricEquals(gauge, 1, ephemeral="0", scope=scope1.name)
        self.assertMetricEquals(gauge, 0, ephemeral="0", scope=scope2.name)
        self.assertMetricEquals(gauge, 0, ephemeral="1", scope=scope1.name)
        self.assertMetricEquals(gauge, 1, ephemeral="1", scope=scope2.name)

    def test_count_tokens(self) -> None:
        self.playground.create_worker_activation_token(enabled=False)
        worker = self.playground.create_worker()  # creates a worker token
        self.playground.create_work_request_token(worker=worker)
        self.playground.create_user_token()
        self.playground.create_user_token()

        gauge = self.collector().count_tokens()
        self.assertEqual(gauge.name, "tokens")
        self.assertMetricEquals(
            gauge,
            1,
            enabled="0",
            user="0",
            worker="0",
            token_type="worker-activation",
        )
        self.assertMetricEquals(
            gauge,
            0,
            enabled="1",
            user="0",
            worker="0",
            token_type="worker-activation",
        )
        self.assertMetricEquals(
            gauge,
            0,
            enabled="0",
            user="0",
            worker="0",
            token_type="work-request",
        )
        self.assertMetricEquals(
            gauge,
            1,
            enabled="1",
            user="0",
            worker="0",
            token_type="work-request",
        )
        self.assertMetricEquals(
            gauge, 0, enabled="0", user="0", worker="1", token_type="worker"
        )
        self.assertMetricEquals(
            gauge, 1, enabled="1", user="0", worker="1", token_type="worker"
        )
        self.assertMetricEquals(
            gauge, 0, enabled="0", user="1", worker="0", token_type="user"
        )
        self.assertMetricEquals(
            gauge, 2, enabled="1", user="1", worker="0", token_type="user"
        )

    def test_count_users(self) -> None:
        self.playground.get_default_user()
        user2 = self.playground.create_user("user2")
        user2.is_active = False
        user2.save()
        gauge = self.collector().count_users()
        self.assertEqual(gauge.name, "users")
        self.assertMetricEquals(gauge, 1, active="1")
        self.assertMetricEquals(gauge, 1, active="0")

    def test_count_user_activity(self) -> None:
        with self.freeze_time() as frozen_time:
            for username, ages in (
                ("user1", (0, 365)),
                ("user2", (4, 5)),
                ("user3", (364,)),
                ("user4", (365 * 2,)),
                ("user5", ()),
            ):
                user = self.playground.create_user(username)
                for days in ages:
                    wr = self.playground.create_workflow(created_by=user)
                    wr.created_at = frozen_time - dt.timedelta(days=days)
                    wr.save()

            gauge = self.collector().count_user_activity()
        self.assertEqual(gauge.name, "user_activity")

        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(gauge, 1, scope=scope, le="1")
        self.assertMetricEquals(gauge, 1, scope=scope, le="3")
        self.assertMetricEquals(gauge, 2, scope=scope, le="7")
        self.assertMetricEquals(gauge, 2, scope=scope, le="14")
        self.assertMetricEquals(gauge, 2, scope=scope, le="30")
        self.assertMetricEquals(gauge, 2, scope=scope, le="90")
        self.assertMetricEquals(gauge, 3, scope=scope, le="365")
        self.assertMetricEquals(gauge, 4, scope=scope, le="+Inf")

    @override_settings(SIGNON_PROVIDERS=[Provider(name="salsa", label="Salsa")])
    def test_count_user_identities_activity(self) -> None:
        with self.freeze_time() as frozen_time:
            for username, issuer, ages in (
                ("user1", "salsa", (0, 365)),
                ("user2", None, (4, 5)),
                ("user3", "salsa", (364,)),
                ("user5", "salsa", ()),
                (None, "salsa", (45,)),
            ):
                user: User | None = None
                subject = "ex-user"
                if username:
                    user = self.playground.create_user(username)
                    subject = username
                if issuer:
                    Identity.objects.create(
                        user=user, issuer=issuer, subject=subject, claims={}
                    )
                for days in ages:
                    wr = self.playground.create_workflow(created_by=user)
                    wr.created_at = frozen_time - dt.timedelta(days=days)
                    wr.save()

            gauge = self.collector().count_user_identities_activity()
        self.assertEqual(gauge.name, "user_identities_activity")

        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="1")
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="3")
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="7")
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="14")
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="30")
        self.assertMetricEquals(gauge, 1, scope=scope, issuer="salsa", le="90")
        self.assertMetricEquals(gauge, 2, scope=scope, issuer="salsa", le="365")
        self.assertMetricEquals(
            gauge, 2, scope=scope, issuer="salsa", le="+Inf"
        )
        self.assertNoMetric(gauge, scope=scope, issuer="", le="1")
        self.assertNoNullLabels(gauge)

    @override_settings(SIGNON_PROVIDERS=[Provider(name="salsa", label="Salsa")])
    def test_count_user_identities(self) -> None:
        user = self.playground.get_default_user()
        Identity.objects.create(
            user=user, issuer="salsa", subject="foo", claims={}
        )
        Identity.objects.create(
            user=None, issuer="salsa", subject="bar", claims={}
        )
        gauge = self.collector().count_user_identities()
        self.assertEqual(gauge.name, "user_identities")
        self.assertMetricEquals(gauge, 1, active="1", issuer="salsa")
        self.assertMetricEquals(gauge, 0, active="0", issuer="salsa")

    def test_count_work_requests(self) -> None:
        self.playground.create_worker_task(task_name="noop", mark_pending=True)
        self.playground.create_signing_task(task_name="noop", mark_running=True)
        self.playground.create_worker_task(
            task_name="noop",
            task_data=NoopData(
                build_architecture="amd64",
                backend=BackendType.UNSHARE,
            ),
            mark_running=True,
        )
        gauge = self.collector().count_work_requests()
        self.assertEqual(gauge.name, "work_requests")
        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(
            gauge,
            1,
            task_type=TaskTypes.WORKER,
            task_name="noop",
            scope=scope,
            status=WorkRequest.Statuses.PENDING,
            build_architecture="",
            backend="",
        )
        self.assertMetricEquals(
            gauge,
            1,
            task_type=TaskTypes.SIGNING,
            task_name="noop",
            scope=scope,
            status=WorkRequest.Statuses.RUNNING,
            build_architecture="",
            backend="",
        )
        self.assertMetricEquals(
            gauge,
            1,
            task_type=TaskTypes.WORKER,
            task_name="noop",
            scope=scope,
            status=WorkRequest.Statuses.RUNNING,
            build_architecture="amd64",
            backend=BackendType.UNSHARE,
        )
        self.assertMetricEquals(
            gauge,
            0,
            task_type=TaskTypes.WORKER,
            task_name="noop",
            scope=scope,
            status=WorkRequest.Statuses.RUNNING,
            build_architecture="amd64",
            backend=BackendType.QEMU,
        )
        self.assertNoNullLabels(gauge)

    def test_count_workers(self) -> None:
        self.playground.create_worker(
            worker_type=WorkerType.EXTERNAL,
            extra_dynamic_metadata={
                "system:native_architecture": "amd64",
                "system:architectures": ["amd64", "i386"],
            },
        )
        worker_pool = self.playground.create_worker_pool("test")
        worker2 = self.playground.create_worker(
            worker_type=WorkerType.EXTERNAL,
            worker_pool=worker_pool,
            extra_dynamic_metadata={
                "system:native_architecture": "i386",
                "system:architectures": ["amd64", "x32"],
            },
        )
        # Overrides dynamic:
        worker2.static_metadata = {"system:native_architecture": "amd64"}
        worker2.save()

        worker3 = self.playground.create_worker(
            worker_type=WorkerType.SIGNING,
        )
        worker3.mark_connected()
        self.playground.create_worker_task(
            task_name="noop", mark_running=True, worker=worker3
        )

        gauge = self.collector().count_workers()
        self.assertEqual(gauge.name, "workers")
        self.assertMetricEquals(
            gauge,
            1,
            connected="0",
            busy="0",
            worker_type=WorkerType.EXTERNAL,
            worker_pool="",
            native_architecture="amd64",
            architecture_amd64="1",
            architecture_i386="1",
            architecture_x32="0",
        )
        self.assertMetricEquals(
            gauge,
            1,
            connected="0",
            busy="0",
            worker_type=WorkerType.EXTERNAL,
            worker_pool=worker_pool.name,
            native_architecture="amd64",
            architecture_amd64="1",
            architecture_i386="0",
            architecture_x32="1",
        )
        self.assertMetricEquals(
            gauge,
            0,
            connected="0",
            busy="0",
            worker_type=WorkerType.EXTERNAL,
            worker_pool="",
            native_architecture="i386",
            architecture_amd64="0",
            architecture_i386="0",
            architecture_x32="0",
        )
        self.assertMetricEquals(
            gauge,
            1,
            connected="1",
            busy="1",
            worker_type=WorkerType.SIGNING,
            worker_pool="",
            native_architecture="",
            architecture_amd64="0",
            architecture_i386="0",
            architecture_x32="0",
        )
        self.assertMetricEquals(
            gauge,
            0,
            connected="0",
            busy="0",
            worker_type=WorkerType.SIGNING,
            worker_pool="",
            native_architecture="",
            architecture_amd64="0",
            architecture_i386="0",
            architecture_x32="0",
        )
        self.assertNoNullLabels(gauge)

    def test_count_worker_pool_runtime(self) -> None:
        with self.freeze_time() as frozen_time:
            worker_pool = self.playground.create_worker_pool("test")
            worker_pool_2 = self.playground.create_worker_pool(
                "test2", provider_account=worker_pool.provider_account
            )

            worker = self.playground.create_worker(worker_pool=worker_pool)
            worker.instance_created_at = frozen_time - dt.timedelta(seconds=120)
            worker.save()
            worker2 = self.playground.create_worker(worker_pool=worker_pool)
            worker2.instance_created_at = frozen_time - dt.timedelta(seconds=60)
            worker2.save()
            # Not part of the pool:
            worker3 = self.playground.create_worker()
            worker3.instance_created_at = frozen_time - dt.timedelta(
                seconds=9999
            )
            worker3.save()

            WorkerPoolStatistics.objects.create(
                worker_pool=worker_pool,
                worker=worker,
                runtime=30,
            )
            WorkerPoolStatistics.objects.create(
                worker_pool=worker_pool,
                runtime=15,
            )
            WorkerPoolStatistics.objects.create(
                worker_pool=worker_pool_2,
                runtime=42,
            )

            gauge = self.collector().count_worker_pool_runtime()
        self.assertEqual(gauge.name, "worker_pool_runtime")
        self.assertMetricEquals(gauge, 120 + 15 + 60 + 30, worker_pool="test")
        self.assertMetricEquals(gauge, 42, worker_pool="test2")

    def test_count_workflow_templates(self) -> None:
        self.playground.create_workflow_template(name="test1", task_name="noop")
        gauge = self.collector().count_workflow_templates()
        self.assertEqual(gauge.name, "workflow_templates")
        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(gauge, 1, task_name="noop", scope=scope)
        self.assertMetricEquals(gauge, 0, task_name="sbuild", scope=scope)

    def test_count_workspaces(self) -> None:
        self.playground.get_default_workspace()
        self.playground.create_workspace(name="private", public=False)
        scope2 = self.playground.get_or_create_scope("scope2")
        self.playground.create_workspace(
            name="expiring",
            expiration_delay=dt.timedelta(days=14),
            scope=scope2,
        )

        gauge = self.collector().count_workspaces()
        self.assertEqual(gauge.name, "workspaces")
        scope = self.playground.get_default_scope().name
        self.assertMetricEquals(gauge, 1, private="0", expires="0", scope=scope)
        self.assertMetricEquals(gauge, 1, private="1", expires="0", scope=scope)
        self.assertMetricEquals(
            gauge, 1, private="0", expires="1", scope=scope2.name
        )
        self.assertMetricEquals(
            gauge, 0, private="0", expires="0", scope=scope2.name
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Assign the correct URL configuration for the requested virtual host."""

from django.conf import settings
from django.http import HttpRequest
from django.utils.deprecation import MiddlewareMixin


class VirtualHostMiddleware(MiddlewareMixin):
    """Assign appropriate URL configuration."""

    @property
    def _archive_fqdns(self) -> list[str]:
        """Hosts used for serving archives."""
        if isinstance(settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN, list):
            return settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN
        else:
            return [settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN]

    def process_request(self, request: HttpRequest) -> None:
        """Process request in middleware."""
        host = request.get_host()
        if host in self._archive_fqdns:
            setattr(request, "urlconf", "debusine.web.archives.urls")
            setattr(request, "is_archive_view", True)
        else:
            setattr(request, "is_archive_view", False)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test Asset data models."""

import base64
from unittest import TestCase, mock

from debusine.assets import (
    AWSProviderAccountConfiguration,
    AWSProviderAccountCredentials,
    AWSProviderAccountData,
    AssetCategory,
    KeyPurpose,
    SigningKeyData,
    asset_categories,
    asset_data_model,
)


class TestBaseAssetDataModel(TestCase):
    """Tests for BaseAssetDataModel."""

    def test_asset_data_model(self) -> None:
        """Test lookup of a known category."""
        data = SigningKeyData(
            purpose=KeyPurpose.UEFI,
            fingerprint="ABC123",
            public_key=base64.b64encode(b"PUBLIC").decode(),
            description="A Test Signing Key",
        )
        model = asset_data_model(AssetCategory.SIGNING_KEY, data.model_dump())
        self.assertEqual(model, data)

    def test_asset_data_model_unknown(self) -> None:
        """Test lookup of an unknown category."""
        with (
            # Replace AssetCategory() with a noop lookup
            mock.patch("debusine.assets.models.AssetCategory", new=str),
            self.assertRaisesRegex(
                ValueError, r"No data model for unknown exists\."
            ),
        ):
            asset_data_model("unknown", {})

    def test_asset_data_model_discriminator(self) -> None:
        data = AWSProviderAccountData(
            name="test",
            configuration=AWSProviderAccountConfiguration(
                region_name="test-region"
            ),
            credentials=AWSProviderAccountCredentials(
                access_key_id="access-key", secret_access_key="secret-key"
            ),
        )
        model = asset_data_model(
            AssetCategory.CLOUD_PROVIDER_ACCOUNT, data.model_dump()
        )
        self.assertEqual(model, data)

    def test_asset_data_model_discriminator_missing(self) -> None:
        with (
            self.assertRaisesRegex(
                ValueError,
                (
                    r"Discriminator key provider_type required for "
                    r"debusine:cloud-provider-account missing in data\."
                ),
            ),
        ):
            asset_data_model(AssetCategory.CLOUD_PROVIDER_ACCOUNT, {})

    def test_asset_data_model_discriminator_unknown(self) -> None:
        with (
            self.assertRaisesRegex(
                ValueError,
                (
                    r"No data model for debusine:cloud-provider-account with "
                    r"provider_type=unknown\."
                ),
            ),
        ):
            asset_data_model(
                AssetCategory.CLOUD_PROVIDER_ACCOUNT,
                {"provider_type": "unknown"},
            )

    def test_asset_categories(self) -> None:
        categories = asset_categories()
        self.assertIn(AssetCategory.CLOUD_PROVIDER_ACCOUNT, categories)
        self.assertIn(AssetCategory.SIGNING_KEY, categories)


class SigningKeyDataTests(TestCase):
    """Test the SigningKeyData model."""

    def test_public_key_base64(self) -> None:
        signing_key_data = SigningKeyData(
            purpose=KeyPurpose.UEFI,
            fingerprint="ABC123",
            public_key=base64.b64encode(b"PUBLIC").decode(),
            description=None,
        )
        self.assertEqual(signing_key_data.public_key, b"PUBLIC")
        self.assertEqual(
            signing_key_data.model_dump()["public_key"],
            base64.b64encode(b"PUBLIC").decode(),
        )

    def test_public_key_not_base64(self) -> None:
        with self.assertRaisesRegex(ValueError, r"Base64 decoding error"):
            SigningKeyData(
                purpose=KeyPurpose.UEFI,
                fingerprint="ABC123",
                public_key="PUBLIC",
                description=None,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine command line interface, work request management commands."""

import argparse
import datetime as dt
import os
import shutil
import subprocess
import tempfile
from collections.abc import Iterable
from itertools import islice
from pathlib import Path
from typing import override

import rich
import rich.json
from rich.table import Table

from debusine.artifacts import Upload
from debusine.artifacts.models import WorkRequestResults
from debusine.client.commands.base import (
    DebusineCommand,
    ModelCommand,
    RequiredInputDataCommand,
    WorkspaceCommand,
)
from debusine.client.debusine import Debusine
from debusine.client.models import (
    WorkRequestExternalDebsignRequest,
    WorkRequestRequest,
    WorkRequestResponse,
)


class WorkRequestModelCommand(
    ModelCommand[WorkRequestResponse], DebusineCommand
):
    """Output WorkRequestResponse objects."""

    @override
    def _list_rich(self, instances: Iterable[WorkRequestResponse]) -> None:
        table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
        table.add_column("ID")
        table.add_column("Task")
        table.add_column("Status")
        table.add_column("Progress")
        table.add_column("Worker")
        table.add_column("Task data")
        for wr in instances:
            url = self._url(wr)
            status = [wr.status]
            if wr.result != WorkRequestResults.NONE:
                status.append(wr.result)
            progress = [f"Created: {self._format_timestamp(wr.created_at)}"]
            if wr.started_at:
                progress.append(
                    f"Started: {self._format_timestamp(wr.started_at)}"
                )
            if wr.completed_at:
                progress.append(
                    f"Completed: {self._format_timestamp(wr.completed_at)}"
                )
            table.add_row(
                f"[link={url}]#{wr.id}[/]",
                f"{wr.task_type} {wr.task_name}",
                " ".join(status),
                "\n".join(progress),
                # TODO: show the worker name, if we can extend the API to get
                # that information
                self._format_int(wr.worker),
                rich.json.JSON.from_data(wr.task_data),
            )
        rich.print(table)

    def _url(self, instance: WorkRequestResponse) -> str:
        return self.debusine.webui_url(instance.get_link_url("webui_self"))

    def _format_timestamp(self, ts: dt.datetime | None) -> str | None:
        if ts is None:
            return None
        return ts.strftime("%Y-%m-%d %H:%M:%S")

    def _format_int(self, value: int | None) -> str | None:
        if value is None:
            return None
        return str(value)

    def _format_float(self, value: float) -> str | None:
        return f"{value:.2f}"

    @override
    def _show_rich(self, instance: WorkRequestResponse) -> None:
        url = self._url(instance)
        table = Table(box=rich.box.SIMPLE, show_header=False)
        table.add_column("Name", justify="right")
        table.add_row("ID:", f"#{instance.id} [link={url}]{url}[/]")
        table.add_row("Task:", f"{instance.task_type} {instance.task_name}")
        table.add_row("Workspace:", f"{instance.scope}/{instance.workspace}")
        table.add_row(
            "Created at:", self._format_timestamp(instance.created_at)
        )
        if instance.started_at is not None:
            table.add_row(
                "Started at:", self._format_timestamp(instance.started_at)
            )
        if instance.completed_at is not None:
            table.add_row(
                "Completed at:", self._format_timestamp(instance.completed_at)
            )
        if instance.duration is not None:
            table.add_row("Duration:", self._format_float(instance.duration))
        if instance.result == WorkRequestResults.NONE:
            table.add_row("Status:", instance.status)
        else:
            table.add_row("Status:", f"{instance.status} {instance.result}")
        if instance.worker is not None:
            # TODO: show the worker name, if we can extend the API to get that
            # information
            table.add_row("Worker:", self._format_int(instance.worker))
        if instance.priority_base != 0 or instance.priority_adjustment != 0:
            table.add_row(
                "Priority:",
                f"{instance.priority_base} + {instance.priority_adjustment}",
            )
        table.add_row(
            "Task data:", rich.json.JSON.from_data(instance.task_data)
        )
        if instance.dynamic_task_data:
            table.add_row(
                "Dynamic task data:",
                rich.json.JSON.from_data(instance.dynamic_task_data),
            )
        if instance.artifacts:
            table.add_row(
                "Artifacts:",
                rich.json.JSON.from_data(instance.artifacts),
            )
        rich.print(table)


class List(WorkRequestModelCommand, DebusineCommand, group="work-request"):
    """List up to limit most recent work requests."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            '--limit',
            type=int,
            metavar="LIMIT",
            default=20,
            help="List the LIMIT most recent work requests",
        )

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            work_requests = list(
                islice(self.debusine.work_request_iter(), self.args.limit)
            )
        self.list(work_requests)


class LegacyList(
    List, name="list-work-requests", deprecated="see `work-request list`"
):
    """List up to limit most recent work requests."""


class Show(DebusineCommand, group="work-request"):
    """Print the status of a work request."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            'work_request_id',
            type=int,
            help='Work request id to show the information',
        )

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            work_request = self.debusine.work_request_get(
                self.args.work_request_id
            )

        artifacts_information = []
        for artifact_id in work_request.artifacts:
            with self._api_call_or_fail():
                artifact_information = self.debusine.artifact_get(artifact_id)
            artifacts_information.append(
                artifact_information.model_dump(mode="json")
            )

        result = work_request.model_dump()
        result["artifacts"] = artifacts_information

        self.print_yaml_output(result)


class LegacyShow(
    Show, name="show-work-request", deprecated="see `work-request show`"
):
    """Print the status of a work request."""


class Create(
    WorkRequestModelCommand,
    RequiredInputDataCommand,
    WorkspaceCommand,
    group="work-request",
):
    """
    Create a work request and schedule the execution.

    Work request is read from stdin in YAML format
    """

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            'task_name', type=str, help='Task name for the work request'
        )
        parser.add_argument(
            "--event-reactions",
            type=argparse.FileType("r"),
            help="File path in YAML format requesting notifications.",
        )

    @override
    def run(self) -> None:
        event_reactions = '{}'
        if self.args.event_reactions:
            event_reactions = self.args.event_reactions.read()
            self.args.event_reactions.close()

        task_name = self.args.task_name
        workspace = self.workspace
        parsed_task_data = self.input_data

        parsed_event_reactions = self._parse_yaml_data(event_reactions)

        work_request = WorkRequestRequest(
            task_name=task_name,
            workspace=workspace,
            task_data=parsed_task_data,
            event_reactions=parsed_event_reactions,
        )

        with self._api_call_or_fail():
            work_request_created = self.debusine.work_request_create(
                work_request
            )

        self.feedback("Work request started:")
        self.show(work_request_created)


class LegacyCreate(
    Create, name="create-work-request", deprecated="see `work-request create`"
):
    """
    Create a work request and schedule the execution.

    Work request is read from stdin in YAML format
    """


class Manage(DebusineCommand, group="work-request"):
    """Manage a work request."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "work_request_id", type=int, help="Work request id to manage"
        )
        parser.add_argument(
            "--set-priority-adjustment",
            type=int,
            metavar="ADJUSTMENT",
            help="Set priority adjustment (positive or negative number)",
        )

    @override
    def run(self) -> None:
        work_request_id = self.args.work_request_id
        priority_adjustment = self.args.set_priority_adjustment
        if priority_adjustment is None:
            self._fail("Error: no changes specified")

        with self._api_call_or_fail():
            self.debusine.work_request_update(
                work_request_id, priority_adjustment=priority_adjustment
            )


class LegacyManage(
    Manage, name="manage-work-request", deprecated="see `work-request manage`"
):
    """Manage a work request."""


class Retry(DebusineCommand, group="work-request"):
    """Retry a failed work request."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "work_request_id", type=int, help="Work request id to retry"
        )

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            self.debusine.work_request_retry(self.args.work_request_id)


class LegacyRetry(
    Retry, name="retry-work-request", deprecated="see `work-request retry`"
):
    """Retry a failed work request."""


class Abort(DebusineCommand, group="work-request"):
    """Abort a failed work request."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "work_request_id", type=int, help="Work request id to retry"
        )

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            self.debusine.work_request_abort(self.args.work_request_id)


class LegacyAbort(
    Abort, name="abort-work-request", deprecated="see `work-request abort`"
):
    """Abort a failed work request."""


class Sign(DebusineCommand, group="work-request"):
    """Provide a work request with an external signature."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "work_request_id",
            type=int,
            help="Work request id that needs a signature",
        )
        parser.add_argument(
            "--local-file",
            "-l",
            type=Path,
            help=(
                "Path to the .changes file to sign, locally. "
                "If not specified, it will be downloaded from the server."
            ),
        )
        parser.add_argument(
            "extra_args",
            nargs="*",
            help=(
                "Additional arguments passed to debsign. "
                "Use a -- to pass options, e.g. -- -kMYKEY."
            ),
        )

    @override
    def run(self) -> None:
        # Find out what kind of work request we're dealing with.
        with self._api_call_or_fail():
            work_request = self.debusine.work_request_get(
                self.args.work_request_id
            )
        match (work_request.task_type, work_request.task_name):
            case "Wait", "externaldebsign":
                self._provide_signature_debsign(
                    work_request, self.args.local_file, self.args.extra_args
                )
            case _:
                self._fail(
                    f"Don't know how to provide signature for "
                    f"{work_request.task_type}/{work_request.task_name} "
                    f"work request"
                )

    def _provide_signature_debsign(
        self,
        work_request: WorkRequestResponse,
        local_file: Path | None,
        debsign_args: list[str],
    ) -> None:
        """Provide a work request with an external signature using `debsign`."""
        if local_file is not None:
            if local_file.suffix != ".changes":
                self._fail(
                    f"--local-file {str(local_file)!r} is not a .changes file."
                )
            if not local_file.exists():
                self._fail(f"--local-file {str(local_file)!r} does not exist.")
        # Get a version of the work request with its dynamic task data
        # resolved.
        with self._api_call_or_fail():
            work_request = self.debusine.work_request_external_debsign_get(
                work_request.id
            )
        assert work_request.dynamic_task_data is not None
        with self._api_call_or_fail():
            unsigned = self.debusine.artifact_get(
                work_request.dynamic_task_data["unsigned_id"]
            )

        if local_file:
            expected_changes = [
                file for file in unsigned.files if file.endswith(".changes")
            ][0]
            if local_file.name not in unsigned.files:
                self._fail(
                    f"{str(local_file)!r} is not part of artifact "
                    f"{work_request.dynamic_task_data['unsigned_id']}. "
                    f"Expecting {expected_changes!r}"
                )

        with tempfile.TemporaryDirectory(
            prefix="debusine-debsign-"
        ) as tmp_name:
            # Get the individual files that need to be signed.
            tmp = Path(tmp_name).resolve()
            for name, file_response in unsigned.files.items():
                path = tmp / name
                assert path.resolve().is_relative_to(tmp)
                assert "/" not in name
                if name.endswith((".changes", ".dsc", ".buildinfo")):
                    if local_file:
                        self._fetch_local_file(
                            local_file.parent / name, path, file_response
                        )
                    else:
                        with self._api_call_or_fail():
                            self.debusine.download_artifact_file(
                                unsigned, name, path
                            )
            # Upload artifacts are guaranteed to have exactly one .changes
            # file.
            [changes_path] = [
                path for path in tmp.iterdir() if path.name.endswith(".changes")
            ]

            # Call debsign.
            subprocess.run(
                ["debsign", "--re-sign", changes_path.name, *debsign_args],
                cwd=tmp,
                check=True,
            )

            # Create a new artifact and upload the files to it.
            signed_local = Upload.create(
                changes_file=changes_path, allow_remote=True
            )
            with self._api_call_or_fail():
                signed_remote = self.debusine.upload_artifact(
                    signed_local, workspace=unsigned.workspace
                )

            # Complete the work request using the signed artifact.
            with self._api_call_or_fail():
                self.debusine.work_request_external_debsign_complete(
                    work_request.id,
                    WorkRequestExternalDebsignRequest(
                        signed_artifact=signed_remote.id
                    ),
                )


class LegacySign(
    Sign, name="provide-signature", deprecated="see `work-request sign`"
):
    """Provide a work request with an external signature."""


class OnCompleted(DebusineCommand, group="work-request", name="on-completed"):
    """
    Execute a command when a work request is completed.

    Arguments to the command: any additional arguments provided
    here, followed by WorkRequest.id and WorkRequest.result
    """

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "--workspace",
            "-w",
            nargs="+",
            type=str,
            help="Workspace names to monitor. If not specified: receive "
            "notifications from all the workspaces that the user "
            "has access to",
        )
        parser.add_argument(
            "--last-completed-at",
            type=Path,
            help="If not passed: does not get any past notifications. "
            "If passed: write into it the WorkRequest.completed_at "
            "and when running again retrieve missed notifications",
        )
        parser.add_argument(
            "command",
            type=str,
            help="Path to the command to execute",
        )
        parser.add_argument(
            "extra_args",
            nargs="*",
            help="Additional arguments to pass to the command to execute",
        )

    @override
    def run(self) -> None:
        workspaces = self.args.workspace
        last_completed_at = self.args.last_completed_at
        command = [self.args.command, *self.args.extra_args]
        if shutil.which(command[0]) is None:
            self._fail(
                f'Error: "{command[0]}" does not exist or is not executable'
            )

        if last_completed_at is not None:
            # Check that the file can be written or created
            if not os.access(last_completed_at.parent, os.W_OK):
                self._fail(
                    f'Error: write access '
                    f'denied for directory "{last_completed_at.parent}"'
                )

            if last_completed_at.exists() and (
                not os.access(last_completed_at, os.W_OK)
                or not os.access(last_completed_at, os.R_OK)
            ):
                self._fail(
                    'Error: write or read access '
                    f'denied for "{last_completed_at}"'
                )

            if not last_completed_at.exists():
                # Create it now. If it fails better now than later
                Debusine.write_last_completed_at(last_completed_at, None)

        self.debusine.on_work_request_completed(
            workspaces=workspaces,
            last_completed_at=last_completed_at,
            command=command,
            working_directory=Path.cwd(),
        )


class LegacyOnCompleted(
    OnCompleted,
    name="on-work-request-completed",
    deprecated="see `work-request on-completed`",
):
    """
    Execute a command when a work request is completed.

    Arguments to the command: any additional arguments provided
    here, followed by WorkRequest.id and WorkRequest.result
    """



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine command line interface, setup command."""

from typing import override

from debusine.client.commands.base import Command


class Setup(Command):
    """Guided setup of debusine-client."""

    @override
    def run(self) -> None:
        from debusine.client.setup import setup_server

        setup_server(
            config_file_path=self.args.config_file,
            server=self.args.server,
            scope=self.args.scope,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Redis backend for prometheus values."""

import datetime as dt
import os
from collections.abc import Callable, Sequence
from typing import Any, Literal, Protocol, cast, override

from prometheus_client.mmap_dict import mmap_key
from prometheus_client.samples import Exemplar

from debusine.server.prometheus_client.redis import (
    _keep_key_from_expiring,
    _key_expiry,
    redis_client,
)

MULTIPROCESS_MODE_T = Literal[
    "all",
    "liveall",
    "min",
    "livemin",
    "max",
    "livemax",
    "sum",
    "livesum",
    "mostrecent",
    "livemostrecent",
    "",
]


class Value(Protocol):
    """Prometheus Client Metric implementation."""

    _multiprocess: bool

    def __init__(
        self,
        typ: str,
        metric_name: str,
        name: str,
        labelnames: Sequence[str],
        labelvalues: Sequence[str],
        help_text: str,
        multiprocess_mode: MULTIPROCESS_MODE_T = "",
        **kwargs: Any,
    ) -> None:
        """Initialize a metric."""

    def inc(self, amount: float) -> None:
        """Increment the metric by amount."""

    def set(self, value: float, timestamp: float | None = None) -> None:
        """Set the metric to value."""

    def get(self) -> float:
        """Get the current metric value."""

    def set_exemplar(self, exemplar: Exemplar) -> None:
        """Set an exemplar value."""
        exemplar  # For vulture

    def get_exemplar(self) -> Exemplar | None:
        """Get any set exemplar value."""


def RedisValue(  # noqa: C901
    process_identifier: Callable[[], str | int] = os.getpid,
) -> type[Value]:
    """
    Value implementation that stores data in a redis/valkey database.

    Key scheme:
    * value:typ:MMAP_KEY

    When a live multiprocess_mode is used, we set the key to expire after
    PROMETHEUS_REDIS_REFRESH_TTL seconds. We launch a daemon thread that
    extends the expiry of all our process' keys every
    PROMETHEUS_REDIS_REFRESH_FREQUENCY.
    """

    class RedisValueImpl(Value):
        _multiprocess: bool = True

        _typ: str
        _metric_name: str
        _name: str
        _labelnames: list[str]
        _labelvalues: list[str]
        _help_text: str
        _multiprocess_mode: MULTIPROCESS_MODE_T
        _expiry: dt.timedelta | None

        _key: str

        def __init__(
            self,
            typ: str,
            metric_name: str,
            name: str,
            labelnames: Sequence[str],
            labelvalues: Sequence[str],
            help_text: str,
            multiprocess_mode: MULTIPROCESS_MODE_T = "",
            **kwargs: Any,
        ) -> None:
            self._typ = typ
            self._metric_name = metric_name
            self._name = name
            self._labelnames = list(labelnames)
            self._labelvalues = list(labelvalues)
            self._help_text = help_text
            self._multiprocess_mode = multiprocess_mode
            self._expiry = None
            if multiprocess_mode:
                if multiprocess_mode in ("mostrecent", "livemostrecent"):
                    raise NotImplementedError(
                        "The 'mostrecent' modes are not supported in RedisValue"
                    )
                assert typ in ("gauge", "gaugehistogram")
                self._labelnames.append("pid")
                self._labelvalues.append("")
                if multiprocess_mode.startswith("live"):
                    self._expiry = _key_expiry()
            self._update_key(True)
            redis_client().set(self._key, 0.0, ex=self._expiry, nx=True)

        def _update_key(self, update: bool = False) -> None:
            if self._multiprocess_mode:
                assert self._labelnames[-1] == "pid"
                new_id = str(process_identifier())
                if new_id != self._labelvalues[-1]:
                    self._labelvalues[-1] = new_id
                    update = True

            if update:
                key = mmap_key(
                    self._metric_name,
                    self._name,
                    self._labelnames,
                    self._labelvalues,
                    self._help_text,
                )
                self._key = f"value:{self._typ}:{self._multiprocess_mode}:{key}"

            if self._expiry and update:
                _keep_key_from_expiring(self._labelvalues[-1], self._key)

        @override
        def inc(self, amount: float) -> None:
            self._update_key()
            client = redis_client()
            client.incrbyfloat(self._key, amount)
            if self._expiry:
                client.expire(self._key, self._expiry)

        @override
        def set(self, value: float, timestamp: float | None = None) -> None:
            timestamp  # for vulture
            self._update_key()
            # TODO: Implement timestamps
            redis_client().set(self._key, value, ex=self._expiry)

        @override
        def get(self) -> float:
            self._update_key()
            # Type bug: https://github.com/redis/redis-py/issues/2399
            value = cast(float | None, redis_client().get(self._key))
            if value is None:
                return 0.0
            return float(value)

        @override
        def set_exemplar(self, exemplar: Exemplar) -> None:
            # TODO: Implement exemplars for redis.
            raise NotImplementedError("Exemplars are not implemented for Redis")

        @override
        def get_exemplar(self) -> Exemplar | None:
            # TODO: Implement exemplars for redis.
            return None

    return RedisValueImpl



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the noop task."""

from unittest import TestCase, mock

import debusine.worker.tags as wtags
from debusine.artifacts.models import WorkRequestResults
from debusine.tasks import Noop, TaskConfigError
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin


class NoopTests(ExternalTaskHelperMixin[Noop], TestCase):
    """Test the Noop class."""

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_get_subject(self) -> None:
        self.assertIsNone(self.get_subject())

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_get_provided_worker_tags(self) -> None:
        task = self.configure_task({"result": WorkRequestResults.SUCCESS})
        self.assertCountEqual(
            task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:noop:version:1",
            ],
        )

    def test_execute_true(self) -> None:
        """Assert execute() return SUCCESS (result is SUCCESS)."""
        noop = self.configure_task({"result": WorkRequestResults.SUCCESS})
        with mock.patch.object(
            noop, "_upload_work_request_debug_logs", autospec=True
        ):
            self.assertEqual(noop.execute(), WorkRequestResults.SUCCESS)

    def test_execute_false(self) -> None:
        """Assert execute() return FAILURE (result is FAILURE)."""
        noop = self.configure_task({"result": WorkRequestResults.FAILURE})
        with mock.patch.object(
            noop, "_upload_work_request_debug_logs", autospec=True
        ):
            self.assertEqual(noop.execute(), WorkRequestResults.FAILURE)

    def test_no_additional_properties(self) -> None:
        """Assert no additional properties."""
        error_msg = "Extra inputs are not permitted"
        with self.assertRaisesRegex(TaskConfigError, error_msg):
            Noop({"input": {}})



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to list users."""

from typing import Any, NoReturn, override

from django.contrib.auth import get_user_model
from django.core.management import CommandParser

from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.management.management_utils import Users


class Command(DebusineBaseCommand):
    """Command to list users."""

    help = "List users sorted by date joined."

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            '--yaml', action="store_true", help="Machine readable YAML output"
        )

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """List users."""
        users = get_user_model().objects.filter(is_system=False)
        Users(options["yaml"]).print(users, self.stdout)
        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to manage a channel notification."""

import argparse
import sys
from typing import Any, NoReturn, override

from django.core.exceptions import ValidationError
from django.core.management import CommandError, CommandParser

from debusine.db.models import NotificationChannel
from debusine.django.management.debusine_base_command import DebusineBaseCommand


class Command(DebusineBaseCommand):
    """Command to manage channel notifications."""

    help = "Manage channel notification"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        subparsers = parser.add_subparsers(dest="action")

        # change name
        change_name = subparsers.add_parser("change-name")
        change_name.add_argument("name", help="Channel to modify")
        change_name.add_argument("new_name", help="New name")

        # change data
        change_data = subparsers.add_parser("change-data")
        change_data.add_argument("name", help="Channel to modify")
        change_data.add_argument(
            "--data",
            type=argparse.FileType("r"),
            help="New data. File path (or - for stdin) to read the data for "
            "the notification. YAML format. Defaults to stdin.",
            default="-",
        )

    @override
    def cleanup_arguments(self, *args: Any, **options: Any) -> None:
        if options["action"] == "change-data" and options["data"] != sys.stdin:
            options["data"].close()

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Manage the notification channel."""
        name = options["name"]
        try:
            notification_channel = NotificationChannel.objects.get(name=name)
        except NotificationChannel.DoesNotExist:
            raise CommandError(
                f'NotificationChannel "{name}" does not exist', returncode=3
            )

        action = options["action"]

        if action == "change-name":
            notification_channel.name = options["new_name"]
            notification_channel.save()
        elif action == "change-data":  # pragma: no cover
            data = options["data"].read()
            notification_channel.data = self.parse_yaml_data(data)
            try:
                notification_channel.save()
            except ValidationError as exc:
                raise CommandError(f"Invalid data: {exc}", returncode=3)

        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Utility functions for tests that require collections."""

from typing import Any

from debusine.artifacts.models import ArtifactCategory
from debusine.db.models import Artifact
from debusine.test.django import TestCase


class CollectionTestMixin(TestCase):
    """Utility functions to create test collections."""

    def create_source_package(
        self, name: str, version: str, dsc_fields: dict[str, Any] | None = None
    ) -> Artifact:
        """Create a minimal `debian:source-package` artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={
                "name": name,
                "version": version,
                "type": "dpkg",
                "dsc_fields": dsc_fields or {},
            },
            paths=[f"{name}_{version}.dsc"],
            create_files=True,
            skip_add_files_in_store=True,
        )
        return artifact
